"""
    read only queries

"""
import concurrent
import logging
import re
from typing import Any, Dict, List

import requests
from component_registry_bindings.bindings.python_client.models import Component

from griffon import (
    COMMUNITY_COMPONENTS_API_URL,
    CORGI_API_URL,
    OSIDB_API_URL,
    CommunityComponentService,
    CorgiService,
    OSIDBService,
)

logger = logging.getLogger("griffon")


class product_stream_summary:
    """retrieve product_stream summary"""

    name = "product_stream_summary"
    description = "retrieve product_stream summary"
    allowed_params = ["strict_name_search", "all", "product_stream_name", "ofuri", "verbose"]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.product_stream_name = self.params.get("product_stream_name")
        self.ofuri = self.params.get("ofuri")
        self.strict_name_search = self.params.get("strict_name_search", None)
        self.all = self.params.get("all", None)

    def execute(self, status=None) -> List[Dict[str, Any]]:
        cond = {}

        if self.ofuri:
            cond["ofuri"] = self.ofuri
        elif not self.strict_name_search:
            if not self.all:
                cond["re_name"] = self.product_stream_name
        else:
            if not self.all:
                cond["name"] = self.product_stream_name

        # TODO - corgi bindings need to support ofuri in product_streams
        product_streams = self.corgi_session.product_streams.retrieve_list(
            **cond,
            limit=400,
            include_fields="link,ofuri,name,products,product_versions,brew_tags,manifest",
        )

        results = []
        if product_streams.results:
            for ps in product_streams.results:
                result = {
                    "link": ps.link,
                    "ofuri": ps.ofuri,
                    "name": ps.name,
                    "product": ps.products[0]["name"],
                    "product_version": str([pv["name"] for pv in ps.product_versions]),
                    "brew_tags": [brew_tag for brew_tag in ps.brew_tags.to_dict().keys()],
                    # "build_count": ps.build_count,
                    "manifest_link": ps.manifest,
                    "latest_components_link": f"{CORGI_API_URL}/api/v1/components?ofuri={ps.ofuri}&view=summary",  # noqa
                    "all_components_link": f"{CORGI_API_URL}/api/v1/components?product_streams={ps.ofuri}&include_fields=link,name,purl",  # noqa
                }
                results.append(result)
        else:
            if "ofuri" in product_streams:
                result = {
                    "link": product_streams["link"],
                    "ofuri": product_streams["ofuri"],
                    "name": product_streams["name"],
                    "product": product_streams["products"][0]["name"],
                    "product_version": product_streams["product_versions"][0]["name"],
                    "brew_tags": "",
                    "manifest_link": product_streams["manifest"],
                    "latest_components_link": f"{CORGI_API_URL}/api/v1/components?ofuri={product_streams['ofuri']}&view=summary",  # noqa
                    "all_components_link": f"{CORGI_API_URL}/api/v1/components?product_streams={product_streams['ofuri']}&include_fields=link,name,purl",  # noqa
                }
                results.append(result)
            else:
                logger.warning("No such product stream")
        return results


class products_versions_affected_by_specific_cve_query:
    """Given a specific CVE ID, what products are affected?"""

    name = "product-versions-affected-by-specific-cve"
    description = "Given a specific CVE ID, what product versions are affected?"
    allowed_params = ["cve_id"]

    def __init__(self, params: dict) -> None:
        self.osidb_session = OSIDBService.create_session()
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.cve_id = self.params.get("cve_id")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")
        self.component_type = self.params.get("component_type")
        self.namespace = self.params.get("namespace")

    def execute(self, status=None) -> dict:
        cond = {}
        if self.affectedness:
            cond["affectedness"] = self.affectedness
        if self.affect_resolution:
            cond["resolution"] = self.affect_resolution
        if self.affect_impact:
            cond["impact"] = self.affect_impact

        component_cond = {}
        if self.namespace:
            component_cond["namespace"] = self.namespace
        if self.component_type:
            component_cond["type"] = self.component_type

        flaw = self.osidb_session.flaws.retrieve(self.cve_id)
        affects = flaw.affects
        results = list()
        product_versions = set()
        product_streams = set()
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for affect in affects:
                futures.append(
                    executor.submit(
                        self.corgi_session.components.retrieve_list,
                        name=affect.ps_component,
                        latest_components_by_streams=True,
                        include_fields="product_streams.name,product_versions.name",
                    )
                )
            for future in concurrent.futures.as_completed(futures):
                try:
                    for c in future.result().results:
                        results.append(c.to_dict())
                except Exception as exc:
                    logger.error("%r generated an exception: %s" % (future, exc))
                    exit(0)

            for c in results:
                for ps in c["product_streams"]:
                    product_streams.add(ps["name"])
                for pv in c["product_versions"]:
                    product_versions.add(ps["name"])
        return {
            "link": f"{OSIDB_API_URL}/osidb/api/v1/flaws/{flaw.cve_id}",
            "cve_id": flaw.cve_id,
            "title": flaw.title,
            "description": flaw.description,
            "product_versions": sorted(list(product_versions)),
            "product_streams": sorted(list(product_streams)),
            # "components": results,
        }


class products_containing_specific_component_query:
    """What products contain a specific component?"""

    name = "products_containing_specific_component_query"
    description = "What products contain a specific component?"
    allowed_params = [
        "component_name",
        "purl",
        "arch",
        "namespace",
        "component_type",
        "strict_name_search",
        "affect_mode",
        "search_latest",
        "search_all",
        "search_all_roots",
        "search_related_url",
        "search_community",
        "search_upstreams",
        "filter_rh_naming",
        "search_redhat",
        "no_community",
        "no_middleware",
        "no_upstream_affects",
        "include_inactive_product_streams",
        "include_product_stream_excluded_components",
        "output_type_filter",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.purl = self.params.get("purl")

    def execute(self, status=None) -> dict:
        c = self.corgi_session.components.retrieve_list(
            purl=self.purl,
        )
        return c["product_streams"]


def async_retrieve_components(
    corgi_session, params, components_initial, component_cnt, item_limit=50
):
    components = list()
    if component_cnt <= item_limit:
        components.extend(components_initial.results)
    elif component_cnt > item_limit:
        components.extend(components_initial.results)
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for batch in range(item_limit, component_cnt, item_limit):
                futures.append(
                    executor.submit(
                        corgi_session.components.retrieve_list,
                        **params,
                        offset=batch,
                        limit=item_limit,  # noqa
                    )
                )
            for future in concurrent.futures.as_completed(futures):
                try:
                    components.extend(future.result().results)
                except Exception as exc:
                    logger.warning("%r generated an exception: %s" % (future, exc))
    return components


class products_containing_component_query:
    """What products contain a component?"""

    name = "products_containing_component_query"
    description = "What products contain a component?"
    allowed_params = [
        "component_name",
        "purl",
        "arch",
        "namespace",
        "component_type",
        "strict_name_search",
        "search_latest",
        "search_all",
        "search_all_roots",
        "search_related_url",
        "search_redhat",
        "search_community",
        "search_upstreams",
        "filter_rh_naming",
        "no_community",
        "no_middleware",
        "no_upstream_affects",
        "include_inactive_product_streams",
        "include_product_stream_excluded_components",
        "output_type_filter",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.component_name = self.params.get("component_name", "")
        self.component_type = self.params.get("component_type", "")
        self.strict_name_search = self.params.get("strict_name_search")
        self.search_deps = self.params.get("search_deps")
        self.ns = self.params.get("namespace")
        self.search_latest = self.params.get("search_latest")
        self.search_all = self.params.get("search_all")
        self.search_all_roots = self.params.get("search_all_roots")
        self.search_related_url = self.params.get("search_related_url")
        self.search_redhat = self.params.get("search_redhat")
        self.search_community = self.params.get("search_community")
        self.search_upstreams = self.params.get("search_upstreams")
        self.filter_rh_naming = self.params.get("filter_rh_naming")
        self.no_community = self.params.get("no_community")
        if not self.no_community:
            self.community_session = CommunityComponentService.create_session()

    def execute(self, status=None) -> List[Dict[str, Any]]:
        status.update("griffoning: searching component-registry.")
        item_limit = 50
        results = []
        params = {
            "include_fields": "link,purl,type,name,related_url,namespace,software_build,nvr,release,version,arch,product_streams.product_versions,product_streams.name,product_streams.ofuri,product_streams.active,product_streams.exclude_components",  # noqa
        }
        # sources.nvr,sources.purl,sources.name,sources.namespace,sources.download_url,sources.related_url
        # upstreams.nvr,upstreams.purl,upstreams.namespace,upstreams.name,upstreams.download_url,upstreams.related_url
        if self.search_latest:
            params["latest_components_by_streams"] = "True"
            if not self.strict_name_search:
                params["re_name"] = self.component_name
            else:
                params["name"] = self.component_name
            if self.ns:
                params["namespace"] = self.ns

            component_initial = self.corgi_session.components.retrieve_list(
                limit=item_limit, **params
            )
            status.update(f"griffoning: found {component_initial.count} latest component(s).")
            latest_components: list = async_retrieve_components(
                self.corgi_session,
                params,
                component_initial,
                component_initial.count,
                item_limit=item_limit,
            )
            results.extend(latest_components)
            if not self.no_community:
                community_component_initial = self.community_session.components.retrieve_list(
                    limit=item_limit, **params
                )
                status.update(
                    f"griffoning: found {community_component_initial.count} latest community component(s)."  # noqa
                )
                community_latest_components: list = async_retrieve_components(
                    self.community_session,
                    params,
                    community_component_initial,
                    community_component_initial.count,
                    item_limit=item_limit,
                )
                results.extend(community_latest_components)

        if self.search_related_url:
            # Note: related_url filter has no concept of strict
            params["related_url"] = self.component_name
            if self.ns:
                params["namespace"] = self.ns
            if self.component_type:
                params["type"] = self.component_type

            component_initial = self.corgi_session.components.retrieve_list(
                limit=item_limit, **params
            )
            status.update(f"griffoning: found {component_initial.count} related url component(s).")
            related_url_components: list = async_retrieve_components(
                self.corgi_session,
                params,
                component_initial,
                component_initial.count,
                item_limit=item_limit,
            )
            results.extend(related_url_components)

        if self.search_all:
            if not self.strict_name_search:
                params["re_name"] = self.component_name
            else:
                params["name"] = self.component_name
            if self.component_type:
                params["type"] = self.component_type
            if self.ns:
                params["namespace"] = self.ns

            all_component_initial = self.corgi_session.components.retrieve_list(
                limit=item_limit, **params
            )
            status.update(f"griffoning: found {all_component_initial.count} all component(s).")
            all_components: list = async_retrieve_components(
                self.corgi_session,
                params,
                all_component_initial,
                all_component_initial.count,
                item_limit=item_limit,
            )
            results.extend(all_components)

        if self.search_all_roots:
            params["type"] = "RPM"
            params["arch"] = "src"
            if not self.strict_name_search:
                params["re_name"] = self.component_name
            else:
                params["name"] = self.component_name
            if self.ns:
                params["namespace"] = self.ns

            all_src_component_initial = self.corgi_session.components.retrieve_list(
                limit=item_limit, **params
            )
            all_src_components: list = async_retrieve_components(
                self.corgi_session,
                params,
                all_src_component_initial,
                all_src_component_initial.count,
                item_limit=item_limit,
            )
            params["type"] = "OCI"
            params["arch"] = "noarch"
            all_noarch_component_initial = self.corgi_session.components.retrieve_list(
                limit=item_limit, **params
            )
            all_noarch_components: list = async_retrieve_components(
                self.corgi_session,
                params,
                all_noarch_component_initial,
                all_noarch_component_initial.count,
                item_limit=item_limit,
            )
            all_root_components = all_src_components + all_noarch_components
            results.extend(all_root_components)

        if self.search_upstreams:
            # Note: upstreams only takes a purl ... so we must use re_upstreams for
            # both strict and not strict search
            params["namespace"] = "UPSTREAM"
            if not self.strict_name_search:
                params["re_name"] = self.component_name
            else:
                params["name"] = self.component_name
            if self.component_type:
                params["type"] = self.component_type

            component_initial = self.corgi_session.components.retrieve_list(
                limit=item_limit, **params
            )
            status.update(f"griffoning: found {component_initial.count} upstream component(s).")
            upstream_components: list = async_retrieve_components(
                self.corgi_session,
                params,
                component_initial,
                component_initial.count,
                item_limit=item_limit,
            )
            results.extend(upstream_components)
            if not self.no_community:
                component_community_initial = self.community_session.components.retrieve_list(
                    limit=item_limit, **params
                )
                status.update(
                    f"griffoning: found {component_community_initial.count} community upstream component(s)."  # noqa
                )
                commmunity_upstream_components: list = async_retrieve_components(
                    self.community_session,
                    params,
                    component_community_initial,
                    component_community_initial.count,
                    item_limit=item_limit,
                )
                results.extend(commmunity_upstream_components)

        if self.filter_rh_naming:
            flags = re.IGNORECASE
            patterns = [
                # binutils
                re.compile(
                    f"(devtoolset\\-[0-9]+\\-|mingw\\-|gcc\\-toolset\\-[0-9]+\\-)?{self.component_name}[0-9\\.]*$",  # noqa
                    flags=flags,
                ),
                # compat-* style
                re.compile(f"(compat\\-)?{self.component_name}[0-9\\.]*(\\-[0-9]+)?$", flags=flags),
                # kernel
                re.compile(f"^{self.component_name}(\\-rt)?$", flags=flags),
                # qemu
                re.compile(f"^{self.component_name}(\\-kvm(\\-rhev|\\-ma)?)?$", flags=flags),
                # webkit
                re.compile(f"^{self.component_name}([0-9])?(gtk)?([0-9])?$", flags=flags),
            ]

            filtered_results = []
            for result in results:
                is_matched = False
                for p in patterns:
                    if is_matched:
                        break
                    if type(result) == Component:
                        m = p.match(result.name)
                        if m:
                            filtered_results.append(result)
                            is_matched = True
                            break
                    else:
                        m = p.match(result["name"])
                        if m:
                            filtered_results.append(result)
                            is_matched = True
                            break

            results = filtered_results

        if not self.no_community and (
            self.search_community or self.search_all or self.search_all_roots
        ):
            params["type"] = "RPM"
            params["arch"] = "src"
            if not self.strict_name_search:
                params["re_name"] = self.component_name
            else:
                params["name"] = self.component_name
            if self.search_upstreams:
                params["namespace"] = "UPSTREAM"
            if self.ns:
                params["namespace"] = self.ns

            if self.component_type:
                params["type"] = self.component_type

            component_initial = self.community_session.components.retrieve_list(
                limit=item_limit, **params
            )
            status.update(
                f"griffoning: found {component_initial.count} community RPM src component(s)."
            )
            commmunity_src_components: list = async_retrieve_components(
                self.community_session,
                params,
                component_initial,
                component_initial.count,
                item_limit=item_limit,
            )
            params["type"] = "OCI"
            params["arch"] = "noarch"
            component_initial_noarch = self.community_session.components.retrieve_list(
                limit=item_limit, **params
            )
            status.update(
                f"griffoning: found {component_initial_noarch.count} community OCI noarch component(s)."  # noqa
            )
            commmunity_noarch_components: list = async_retrieve_components(
                self.community_session,
                params,
                component_initial_noarch,
                component_initial_noarch.count,
                item_limit=item_limit,
            )
            community_components = commmunity_src_components + commmunity_noarch_components
            results.extend(community_components)

        return results


class components_containing_specific_component_query:
    """What components contain a specific component?"""

    name = "components_containing_specific_component_query"
    description = "What components contain a specific component?"
    allowed_params = [
        "component_re_name",
        "component_name",
        "purl",
        "component_type",
        "component_version",
        "component_arch",
        "namespace",
        "strict_name_search",
        "verbose",
    ]

    def __init__(self, params: dict):
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.purl = self.params.get("purl")

    def execute(self, status=None) -> dict:
        if self.purl:
            c = self.corgi_session.components.retrieve_list(
                purl=self.purl,
            )
            component_type = self.params["component_type"]
            sources = c["sources"]
            if component_type:
                sources = [source for source in sources if component_type.lower() in source["purl"]]
        return {
            "link": c["link"],
            "type": c["type"],
            "name": c["name"],
            "version": c["version"],
            "purl": c["purl"],
            "sources": sources,
        }


class components_containing_component_query:
    """What components contain a component?"""

    name = "components_containing_component_query"
    description = "What components contain a component?"
    allowed_params = [
        "component_name",
        "purl",
        "component_type",
        "component_version",
        "component_arch",
        "namespace",
        "strict_name_search",
        "verbose",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.params = params
        self.component_name = self.params.get("component_name")
        self.component_type = self.params.get("component_type")
        self.component_version = self.params.get("component_version")
        self.component_arch = self.params.get("component_arch")
        self.namespace = self.params.get("namespace")
        self.strict_name_search = self.params.get("strict_name_search")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        cond = {}
        if not self.strict_name_search:
            cond["re_name"] = self.component_name
        else:
            cond["name"] = self.component_name
        if self.component_version:
            cond["version"] = self.component_version
        if self.component_arch:
            cond["arch"] = self.component_arch
        if self.namespace:
            cond["namespace"] = self.namespace

        components: List[Any] = []
        logger.debug("starting parallel http requests")
        component_cnt = self.corgi_session.components.retrieve_list(**cond).count
        if component_cnt < 3000000:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = []
                components = list()
                for batch in range(0, component_cnt, 120):
                    futures.append(
                        executor.submit(
                            self.corgi_session.components.retrieve_list,
                            **cond,
                            offset=batch,
                            include_fields="link,name,type,arch,version,purl,nvr,sources,related_url,download_url",  # noqa
                            limit=120,
                        )
                    )
                for future in concurrent.futures.as_completed(futures):
                    try:
                        components.extend(future.result().results)
                    except Exception as exc:
                        logger.warning("%r generated an exception: %s" % (future, exc))

        results = []
        for c in components:
            sources = [{"link": source["link"], "purl": source["purl"]} for source in c.sources]
            if self.component_type:
                sources = [
                    source for source in sources if self.component_type.lower() in source["purl"]
                ]
            results.append(
                {
                    "link": c.link,
                    "type": c.type,
                    "name": c.name,
                    "version": c.version,
                    "nvr": c.nvr,
                    "arch": c.arch,
                    "download_url": c.download_url,
                    "related_url": c.related_url,
                    "purl": c.purl,
                    "sources": sources,
                }
            )
        return results


class components_affected_by_specific_cve_query:
    """Given a specific CVE ID, what components are affected?"""

    name = "components_affected_by_specific_cve_query"
    description = "Given a CVE ID, what components are affected?"
    allowed_params = [
        "cve_id",
        "affectedness",
        "affect_resolution",
        "affect_impact",
        "component_type",
        "namespace",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params
        self.cve_id = self.params.get("cve_id")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")
        self.component_type = self.params.get("component_type")
        self.namespace = self.params.get("namespace")

    def execute(self, status=None) -> dict:
        cond = {}
        if self.affectedness:
            cond["affectedness"] = self.affectedness
        if self.affect_resolution:
            cond["resolution"] = self.affect_resolution
        if self.affect_impact:
            cond["impact"] = self.affect_impact

        component_cond = {}
        if self.namespace:
            component_cond["namespace"] = self.namespace
        if self.component_type:
            component_cond["type"] = self.component_type

        flaw = self.osidb_session.flaws.retrieve(self.cve_id)
        affects = flaw.affects
        results = list()
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for affect in affects:
                futures.append(
                    executor.submit(
                        self.corgi_session.components.retrieve_list,
                        name=affect.ps_component,
                        latest_components_by_streams=True,
                        include_fields="purl,product_streams,product_versions,software_build",
                    )
                )
            for future in concurrent.futures.as_completed(futures):
                try:
                    for c in future.result().results:
                        results.append(c.to_dict())
                except Exception as exc:
                    logger.error("%r generated an exception: %s" % (future, exc))
                    exit(0)

        return {
            "link": f"{OSIDB_API_URL}/osidb/api/v1/flaws/{flaw.cve_id}",
            "cve_id": flaw.cve_id,
            "title": flaw.title,
            "description": flaw.description,
            "components": results,
        }


class cves_for_specific_component_query:
    """CVEs affecting a specific component?"""

    # include count
    name = "cves_for_specific component"
    description = "Which CVEs affect a specific component ?"
    allowed_params = [
        "component_name",
        "purl",
        "flaw_state",
        "flaw_impact",
        "flaw_resolution",
        "affectedness",
        "affect_resolution",
        "affect_impact",
        "strict_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params
        self.component_name = self.params.get("component_name")
        self.purl = self.params.get("purl")
        self.flaw_state = self.params.get("flaw_state")
        self.flaw_impact = self.params.get("flaw_impact")
        self.flaw_resolution = self.params.get("flaw_resolution")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        components = []
        if self.component_name:
            affects: list = []
            params = {
                "include_fields": "cve_id,title,state,resolution,impact,affects",
            }
            params["affects__ps_component"] = self.component_name
            if self.flaw_state:
                params["state"] = self.flaw_state
            if self.flaw_resolution:
                params["resolution"] = self.flaw_resolution
            if self.flaw_impact:
                params["impact"] = self.flaw_impact
            if self.affectedness:
                params["affects__affectedness"] = self.affectedness
            if self.affect_resolution:
                params["affects__resolution"] = self.affect_resolution
            if self.affect_impact:
                params["affects__impact"] = self.affect_impact

            res = requests.get(f"{OSIDB_API_URL}/osidb/api/v1/flaws", params=params)
            flaws = res.json()
            flaws_cnt = int(flaws["count"])
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = []
                flaws = list()
                for batch in range(0, flaws_cnt, 75):
                    params["offset"] = batch  # type: ignore
                    params["limit"] = 75  # type: ignore
                    futures.append(
                        executor.submit(
                            requests.get,
                            f"{OSIDB_API_URL}/osidb/api/v1/flaws",
                            params=params,
                        )
                    )
                for future in concurrent.futures.as_completed(futures):
                    try:
                        flaws.extend(future.result().json()["results"])
                    except Exception as exc:
                        logger.warning("%r generated an exception: %s" % (future, exc))

                for flaw in flaws:
                    for affect in flaw["affects"]:
                        if self.affectedness:
                            if self.affectedness != affect["affectedness"]:
                                continue
                        if self.affect_resolution:
                            if self.affect_resolution != affect["resolution"]:
                                continue
                        if self.affect_impact:
                            if self.affect_impact != affect["impact"]:
                                continue
                        affects.append(
                            {
                                "link_affect": f"{OSIDB_API_URL}/osidb/api/v1/affects/{affect['uuid']}",  # noqa
                                "link_cve": f"{OSIDB_API_URL}/osidb/api/v1/flaws/{flaw['cve_id']}",  # noqa
                                "link_component": f"{CORGI_API_URL}/api/v1/components?name={affect['ps_component']}&latest_components_by_streams=True",  # noqa
                                "link_community_component": f"{COMMUNITY_COMPONENTS_API_URL}/api/v1/components?name={affect['ps_component']}&latest_components_by_streams=True",  # noqa
                                "flaw_cve_id": flaw["cve_id"],
                                "title": flaw["title"],
                                "flaw_state": flaw["state"],
                                "flaw_resolution": flaw["resolution"],
                                "affect_component_name": affect["ps_component"],
                                "affect_product_version": affect["ps_module"],
                                "affect_affectedness": affect["affectedness"],
                                "affect_impact": affect["impact"],
                                "affect_resolution": affect["resolution"],
                            }
                        )
                components.append(
                    {
                        "link": f"{CORGI_API_URL}/api/v1/components?name={self.component_name}",
                        "name": self.component_name,
                        "affects": affects,
                    }
                )

        if self.purl:
            pass

        return components


class cves_for_specific_product_query:
    name = "cves-for-product"
    description = "What cves affect a specific product ?"
    allowed_params = [
        "product_version_name",
        "ofuri",
        "flaw_state",
        "flaw_impact",
        "flaw_resolution",
        "affectedness",
        "affect_resolution",
        "affect_impact",
        "strict_name_search",
    ]

    def __init__(self, params: dict) -> None:
        self.corgi_session = CorgiService.create_session()
        self.osidb_session = OSIDBService.create_session()
        self.params = params
        self.product_version_name = self.params.get("product_version_name")
        self.ofuri = self.params.get("ofuri")
        self.flaw_state = self.params.get("flaw_state")
        self.flaw_impact = self.params.get("flaw_impact")
        self.flaw_resolution = self.params.get("flaw_resolution")
        self.affectedness = self.params.get("affectedness")
        self.affect_resolution = self.params.get("affect_resolution")
        self.affect_impact = self.params.get("affect_impact")

    def execute(self, status=None) -> List[Dict[str, Any]]:
        components = []
        if self.product_version_name:
            affects: list = []
            params = {
                "include_fields": "cve_id,title,state,resolution,impact,affects",
            }
            params["affects__ps_module"] = self.product_version_name
            if self.flaw_state:
                params["state"] = self.flaw_state
            if self.flaw_resolution:
                params["resolution"] = self.flaw_resolution
            if self.flaw_impact:
                params["impact"] = self.flaw_impact
            if self.affectedness:
                params["affects__affectedness"] = self.affectedness
            if self.affect_resolution:
                params["affects__resolution"] = self.affect_resolution
            if self.affect_impact:
                params["affects__impact"] = self.affect_impact

            res = requests.get(f"{OSIDB_API_URL}/osidb/api/v1/flaws", params=params)
            flaws = res.json()
            flaws_cnt = int(flaws["count"])
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = []
                flaws = list()
                for batch in range(0, flaws_cnt, 75):
                    params["offset"] = batch  # type: ignore
                    params["limit"] = 75  # type: ignore
                    futures.append(
                        executor.submit(
                            requests.get,
                            f"{OSIDB_API_URL}/osidb/api/v1/flaws",
                            params=params,
                        )
                    )
                for future in concurrent.futures.as_completed(futures):
                    try:
                        flaws.extend(future.result().json()["results"])
                    except Exception as exc:
                        logger.warning("%r generated an exception: %s" % (future, exc))

                for flaw in flaws:
                    for affect in flaw["affects"]:
                        if self.affectedness:
                            if self.affectedness != affect["affectedness"]:
                                continue
                        if self.affect_resolution:
                            if self.affect_resolution != affect["resolution"]:
                                continue
                        if self.affect_impact:
                            if self.affect_impact != affect["impact"]:
                                continue
                        affects.append(
                            {
                                "link_affect": f"{OSIDB_API_URL}/osidb/api/v1/affects/{affect['uuid']}",  # noqa
                                "link_cve": f"{OSIDB_API_URL}/osidb/api/v1/flaws/{flaw['cve_id']}",  # noqa
                                "link_component": f"{CORGI_API_URL}/api/v1/components?name={affect['ps_component']}&latest_components_by_streams=True",  # noqa
                                "link_community_component": f"{COMMUNITY_COMPONENTS_API_URL}/api/v1/components?name={affect['ps_component']}&latest_components_by_streams=True",  # noqa
                                "flaw_cve_id": flaw["cve_id"],
                                "title": flaw["title"],
                                "flaw_state": flaw["state"],
                                "flaw_resolution": flaw["resolution"],
                                "affect_component_name": affect["ps_component"],
                                "affect_product_version": affect["ps_module"],
                                "affect_affectedness": affect["affectedness"],
                                "affect_impact": affect["impact"],
                                "affect_resolution": affect["resolution"],
                            }
                        )
                components.append(
                    {
                        "link": f"{CORGI_API_URL}/api/v1/product_versions?name={self.product_version_name}",  # noqa
                        "name": self.product_version_name,
                        "affects": affects,
                    }
                )

        return components



"""
    griffon cli

"""
import configparser
import logging
import os
from configparser import ConfigParser
from functools import partial, wraps

import component_registry_bindings
import osidb_bindings
from osidb_bindings.bindings.python_client.models import Affect, Flaw, Tracker
from pkg_resources import resource_filename  # type: ignore
from rich.logging import RichHandler

from griffon.output import console

__version__ = "0.2.21"

if "CORGI_API_URL" not in os.environ:
    print("Must set CORGI_API_URL environment variable.")
    exit(1)
CORGI_API_URL = os.environ["CORGI_API_URL"]

if "OSIDB_API_URL" not in os.environ:
    print("Must set OSIDB_API_URL environment variable.")
    exit(1)
OSIDB_API_URL = os.environ["OSIDB_API_URL"]
OSIDB_USERNAME = os.getenv("OSIDB_USERNAME", "")
OSIDB_PASSWORD = os.getenv("OSIDB_PASSWORD", "")
OSIDB_AUTH_METHOD = os.getenv("OSIDB_AUTH_METHOD", "kerberos")

# required to enable --search-community
COMMUNITY_COMPONENTS_API_URL = os.getenv(
    "COMMUNITY_COMPONENTS_API_URL", "https://component-registry.fedoraproject.org"
)

# TODO: temporary hack required to enable searching of middleware
MIDDLEWARE_CLI = os.getenv("GRIFFON_MIDDLEWARE_CLI")

GRIFFON_CONFIG_DIR = os.getenv("GRIFFON_API_URL", "~/.griffon")
GRIFFON_RC_FILE = "~/.griffonrc"
GRIFFON_DEFAULT_LOG_FILE = os.getenv("GRIFFON_DEFAULT_LOG_FILE", "~/.griffon/history.log")

logger = logging.getLogger("griffon")

RELATED_MODELS_MAPPING = {Flaw: {"affects": Affect}, Affect: {"trackers": Tracker}}


def config_logging(level="INFO"):
    # if set to 'DEBUG' then we want all the http conversation
    if level == "DEBUG":
        try:
            import http.client as http_client
        except ImportError:
            # Python 2
            import httplib as http_client
        http_client.HTTPConnection.debuglevel = 1

    message_format = "%(asctime)s %(name)s %(levelname)s %(message)s"
    logging.basicConfig(
        level=level, format=message_format, datefmt="[%X]", handlers=[RichHandler()]
    )


def get_config():
    """read ~/.griffonrc ini file, if it does not exist then return some default config"""
    if not os.path.exists(os.path.expanduser(GRIFFON_RC_FILE)):
        default_griffonrc = resource_filename(__name__, "static/default_griffonrc")
        config = configparser.ConfigParser(allow_no_value=True)
        config.read(default_griffonrc)
        return config
    config = ConfigParser()
    config.read(os.path.expanduser(GRIFFON_RC_FILE))
    return config


griffon_config = get_config()


def get_config_option(section, option, default_value=None):
    if griffon_config.has_option(section, option):
        return griffon_config.get(section, option)
    return default_value


def list_config_sections():
    return griffon_config.sections()


class CorgiService:
    name = "component-registry"
    description = "Red Hat component registry"
    has_binding = True

    @staticmethod
    def create_session():
        """init corgi session"""
        try:
            return component_registry_bindings.new_session(
                component_registry_server_uri=CORGI_API_URL
            )
        except:  # noqa
            console.log(f"{CORGI_API_URL} is not accessible.")
            exit(1)

    @staticmethod
    def get_component_types():
        """get component type enum"""
        return (
            component_registry_bindings.bindings.python_client.models.component_type_enum.ComponentTypeEnum  # noqa
        )

    @staticmethod
    def get_component_namespaces():
        """get component namespaces enum"""
        return (
            component_registry_bindings.bindings.python_client.models.namespace_enum.NamespaceEnum
        )

    @staticmethod
    def get_component_arches():
        """get component arch enum"""
        return [
            "src",
            "noarch",
            "i386",
            "ia64",
            "s390",
            "x86_64",
            "s390x",
            "ppc",
            "ppc64",
            "aarch64",
            "ppc64le",
        ]

    @staticmethod
    def get_fields(model, prefix=""):
        """
        get model fields and fields of its related models with
        respective prefixes using dot notation

        eg. field, related_model.field, related_model_1.related_model2.field
        """

        # get rid of the self attribute
        fields = [f"{prefix}{field}" for field in model.get_fields().keys()]
        for name, related_model in RELATED_MODELS_MAPPING.get(model, {}).items():
            fields.extend(CorgiService.get_fields(related_model, prefix=f"{prefix}{name}."))

        return fields


class OSIDBService:
    name = "osidb"
    description = "Open Source Incident database"
    has_binding = True

    @staticmethod
    def create_session():
        """init osidb session"""
        try:
            credentials = {}
            if OSIDB_AUTH_METHOD == "credentials":
                credentials["username"] = OSIDB_USERNAME
                credentials["password"] = OSIDB_PASSWORD
            return osidb_bindings.new_session(osidb_server_uri=OSIDB_API_URL, **credentials)
        except:  # noqa
            console.log(f"{OSIDB_API_URL} is not accessible (or krb ticket has expired).")
            exit(1)

    @staticmethod
    def get_flaw_states():
        """get flaw states enum"""
        return osidb_bindings.bindings.python_client.models.FlawClassificationState

    @staticmethod
    def get_flaw_resolutions():
        """get flaw resolution enum"""
        # TODO: FlawResolutionEnum changed to Resolution01FEnum in OSIDB schema
        # due to some weird drf-spectacular naming clash resolution, there is
        # a way ho to set this to a immutable name however this would require
        # new OSIDB release, so importing the current path and once OSIDB schema
        # is changed and released we can change it back to normal
        #
        # return osidb_bindings.bindings.python_client.models.FlawResolutionEnum
        return osidb_bindings.bindings.python_client.models.ResolutionEnum

    @staticmethod
    def get_flaw_impacts():
        """get flaw impacts enum"""
        return osidb_bindings.bindings.python_client.models.ImpactEnum

    @staticmethod
    def get_affect_affectedness():
        """get affect affectedness enum"""
        return osidb_bindings.bindings.python_client.models.AffectednessEnum

    @staticmethod
    def get_affect_resolution():
        """get affect affectedness enum"""
        # TODO: AffectResolutionEnum changed to Resolution3AcEnum in OSIDB schema
        # due to some weird drf-spectacular naming clash resolution, there is
        # a way ho to set this to a immutable name however this would require
        # new OSIDB release, so importing the current path and once OSIDB schema
        # is changed and released we can change it back to normal
        #
        # return osidb_bindings.bindings.python_client.models.AffectResolutionEnum
        return osidb_bindings.bindings.python_client.models.ResolutionEnum

    @staticmethod
    def get_affect_impact():
        """get affect impact enum"""
        return osidb_bindings.bindings.python_client.models.ImpactEnum

    @staticmethod
    def get_flaw_meta_type():
        """get flaw meta type enum"""
        return osidb_bindings.bindings.python_client.models.MetaTypeEnum

    @staticmethod
    def get_fields(model, prefix=""):
        """
        get model fields and fields of its related models with
        respective prefixes using dot notation

        eg. field, related_model.field, related_model_1.related_model2.field
        """

        # get rid of the self attribute
        fields = [f"{prefix}{field}" for field in model.get_fields().keys()]
        for name, related_model in RELATED_MODELS_MAPPING.get(model, {}).items():
            fields.extend(OSIDBService.get_fields(related_model, prefix=f"{prefix}{name}."))

        return fields

    @staticmethod
    def get_meta_attr_fields(model, prefix=""):
        """
        get model meta attr keys and keys of its related models meta attr with
        respective prefixes using dot notation

        eg. key, related_model.key, related_model_1.related_model2.key
        """
        model_meta_attr = model.get_fields().get("meta_attr")
        if model_meta_attr is None:
            return []

        # get rid of the self attribute and add additional wildcard
        fields = [f"{prefix}{field}" for field in model_meta_attr.get_fields().keys() | {"*"}]
        for name, related_model in RELATED_MODELS_MAPPING.get(model, {}).items():
            fields.extend(
                OSIDBService.get_meta_attr_fields(related_model, prefix=f"{prefix}{name}.")
            )

        return fields


class CommunityComponentService:
    name = "community-component-registry"
    description = "Red Hat component registry"
    has_binding = True

    @staticmethod
    def create_session():
        """init corgi session"""
        try:
            return component_registry_bindings.new_session(
                component_registry_server_uri=COMMUNITY_COMPONENTS_API_URL
            )
        except:  # noqa
            console.log(f"{COMMUNITY_COMPONENTS_API_URL} is not accessible.")
            exit(1)

    @staticmethod
    def get_component_types():
        """get component type enum"""
        return (
            component_registry_bindings.bindings.python_client.models.component_type_enum.ComponentTypeEnum  # noqa
        )

    @staticmethod
    def get_component_namespaces():
        """get component namespaces enum"""
        return (
            component_registry_bindings.bindings.python_client.models.namespace_enum.NamespaceEnum
        )

    @staticmethod
    def get_component_arches():
        """get component arch enum"""
        return [
            "src",
            "noarch",
            "i386",
            "ia64",
            "s390",
            "x86_64",
            "s390x",
            "ppc",
            "ppc64",
            "aarch64",
            "ppc64le",
        ]

    @staticmethod
    def get_fields(model, prefix=""):
        """
        get model fields and fields of its related models with
        respective prefixes using dot notation

        eg. field, related_model.field, related_model_1.related_model2.field
        """

        # get rid of the self attribute
        fields = [f"{prefix}{field}" for field in model.get_fields().keys()]
        for name, related_model in RELATED_MODELS_MAPPING.get(model, {}).items():
            fields.extend(CorgiService.get_fields(related_model, prefix=f"{prefix}{name}."))

        return fields


def progress_bar(
    func=None,
):
    """progress bar decorator"""
    if not func:
        return partial(progress_bar)

    @wraps(func)
    def wrapper(*args, **kwargs):
        obj: dict = args[0].obj
        if obj.get("NO_PROGRESS_BAR"):
            func(*args, **kwargs)
        else:
            with console.status("griffoning", spinner="line"):
                func(*args, **kwargs)

    return wrapper


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    print(f"Version {__version__}")
    ctx.exit()



import pytest
from click.testing import CliRunner

from griffon.cli import cli

pytestmark = pytest.mark.unit


def test_cli_flaws():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "osidb", "flaws"])
    assert result.exit_code == 0


def test_cli_affects():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "osidb", "affects"])
    assert result.exit_code == 0


def test_cli_trackers():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "osidb", "trackers"])
    assert result.exit_code == 0


def test_cli_components():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "component-registry", "components"])
    assert result.exit_code == 0


def test_cli_builds():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "component-registry", "builds"])
    assert result.exit_code == 0


def test_cli_product_streams():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "component-registry", "product-streams"])
    assert result.exit_code == 0


def test_cli_community_components():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "community-component-registry", "components"])
    assert result.exit_code == 0


def test_cli_community_builds():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "community-component-registry", "builds"])
    assert result.exit_code == 0


def test_cli_community_product_streams():
    runner = CliRunner()
    result = runner.invoke(cli, ["entities", "community-component-registry", "product-streams"])
    assert result.exit_code == 0



"""
    Gather up all of the messy 'presentation' logic into one place

"""
import enum
import json
import logging
import re

import click
from packageurl import PackageURL
from rich.console import Console
from rich.text import Text
from rich.tree import Tree

console = Console(color_system="auto")

logger = logging.getLogger("griffon")


class OUTPUT_FORMAT(enum.Enum):
    JSON = "json"
    TEXT = "text"
    TABLE = "table"


class DEST(enum.Enum):
    CONSOLE = "console"
    FILE = "file"


def raw_json_transform(data, show_count: bool) -> dict:
    """normalise all data to dict"""
    if type(data) is list:
        results = []
        for d in data:
            if type(d) is dict:
                results.append(d)
            else:
                results.append(d.to_dict())
        output = {
            "results": results,
        }
        if show_count:
            output["count"] = len(results)  # type: ignore
    else:
        if type(data) is dict:
            output = data
        else:
            output = data.to_dict()
    return output


def entity_type(data):
    entity_type = "unknown"
    if "cve_id" in data:
        entity_type = "flaw"
    if "purl" in data:
        entity_type = "component"
    if "affectedness" in data:
        entity_type = "affect"
    if "ofuri" in data:
        entity_type = "product"
    return entity_type


def component_type_style(type):
    from griffon import CorgiService

    types = [type.value for type in CorgiService.get_component_types()]
    colors = (
        "cornflower_blue",
        "red",
        "dark_slate_gray1",
        "magenta",
        "navy_blue",
        "green1",
        "blue",
        "dark_orange",
        "deep_pink1",
        "red1",
        "green1",
    )
    color = colors[types.index(type)]
    return f"[{color}]{type}[/{color}]"


def output_version(ctx, version):
    if version:
        if version.startswith("sha256") and ctx.obj["SHORT_VERSION_VALUES"]:
            return f"sha256...{version[-8:]}"
    return version


def walk_component_tree(obj, key, tree, show_purl=False):
    if isinstance(obj, dict):
        for k, v in obj.items():
            if isinstance(v, (dict, list)):
                walk_component_tree(v, k, tree, show_purl=show_purl)
    elif isinstance(obj, list):
        for item in obj:
            label = f"{item['node_type']}: {item['nvr']},{item['type']}"
            if show_purl:
                label = f"{item['node_type']}: {item['purl']}"
            child = tree.add(label)
            walk_component_tree(item, None, child, show_purl=show_purl)
    return tree


def text_output_tree(ctx, output, no_wrap=False):
    tree = Tree(
        "component dependency tree",
        guide_style="bold magenta",
    )
    console.print(walk_component_tree(output, None, tree, show_purl=ctx.params["show_purl"]))
    ctx.exit()


def text_output_product_summary(ctx, output, format, exclude_products, no_wrap=False):
    ordered_results = sorted(output["results"], key=lambda d: d["name"])

    if exclude_products:
        exclude_products_results = []
        for result in ordered_results:
            if not any([re.search(match, result["product_version"]) for match in exclude_products]):
                exclude_products_results.append(result)
        ordered_results = exclude_products_results

    if ctx.obj["VERBOSE"] == 0:
        for item in ordered_results:
            console.print(
                Text(item["product"], style="bold magenta u"),
                Text(item["product_version"], style="magenta"),
                Text(item["name"], style="white"),
                f"{item['brew_tags']}",
                no_wrap=no_wrap,
            )
    if ctx.obj["VERBOSE"] == 1:
        for item in ordered_results:
            console.print(
                Text(item["product"], style="bold magenta u"),
                Text(item["product_version"], style="magenta"),
                Text(item["name"], style="white"),
                item["brew_tags"],
                item["ofuri"],
                no_wrap=no_wrap,
            )
    if ctx.obj["VERBOSE"] > 1:
        for item in ordered_results:
            console.print(
                Text(item["product"], style="bold magenta u"),
                Text(item["product_version"], style="magenta"),
                Text(item["name"], style="white"),
                item["brew_tags"],
                item["ofuri"],
                item["manifest_link"],
                no_wrap=no_wrap,
            )
    ctx.exit()


def generate_normalised_results(
    output,
    exclude_products,
    exclude_components,
    output_type_filter,
    include_inactive_product_streams,
    include_product_stream_excluded_components,
):
    normalised_results = list()
    if "results" in output:
        for item in output["results"]:
            for ps in item["product_streams"]:
                # only include component from active product stream
                if ps.get("active") or include_inactive_product_streams:
                    # .griffonrc defined exclude product streams
                    if not any(
                        [
                            re.search(match, ps["product_versions"][0]["name"])
                            for match in exclude_products
                        ]
                    ):
                        # product stream defined exclude components
                        if (
                            not any(
                                [
                                    re.search(match, item["name"])
                                    for match in ps.get("exclude_components", [])
                                ]
                            )
                            or include_product_stream_excluded_components
                        ):
                            # .griffonrc defined exclude components
                            if not any(
                                [re.search(match, item["name"]) for match in exclude_components]
                            ):
                                c = {
                                    "product_version": ps["product_versions"][0]["name"],
                                    "product_stream": ps.get("name"),
                                    "product_stream_active": ps.get("active"),
                                    "namespace": item.get("namespace"),
                                    "name": item.get("name"),
                                    "nvr": item.get("nvr"),
                                    "type": item.get("type"),
                                    "arch": item.get("arch"),
                                    "version": item.get("version"),
                                    "related_url": item.get("related_url"),
                                    "purl": item.get("purl"),
                                    "sources": item.get("sources"),
                                    "upstreams": item.get("upstreams"),
                                }
                                if "software_build" in item:
                                    c["build_source_url"] = item["software_build"].get("source")
                                # output type filter
                                if output_type_filter is None:
                                    normalised_results.append(c)
                                if item.get("type") == output_type_filter:
                                    normalised_results.append(c)
    return normalised_results


def generate_result_tree(normalised_results):
    product_versions = sorted(list(set([item["product_version"] for item in normalised_results])))
    result_tree = {}
    for pv in product_versions:
        result_tree[pv] = {}
        product_streams = sorted(
            list(
                set(
                    [
                        item["product_stream"]
                        for item in normalised_results
                        if item["product_version"] == pv
                    ]
                )
            )
        )
        for ps in product_streams:
            result_tree[pv][ps] = {}
            component_names = sorted(
                list(
                    set(
                        [
                            item["name"]
                            for item in normalised_results
                            if item["product_stream"] == ps
                        ]
                    )
                )
            )
            for cn in component_names:
                result_tree[pv][ps][cn] = {}
                nvrs = [
                    item
                    for item in normalised_results
                    if item["product_stream"] == ps and item["name"] == cn
                ]

                for nvr in nvrs:
                    result_tree[pv][ps][cn][nvr["nvr"]] = nvr
    return result_tree


def generate_affects(
    ctx, result_tree, exclude_components, flaw_operation, format="text", no_wrap=False
):
    search_component_name = ctx.params["component_name"]
    for pv in result_tree.keys():
        component_names = set()
        for ps in result_tree[pv].keys():
            for component_name in result_tree[pv][ps].keys():
                for nvr in result_tree[pv][ps][component_name].keys():
                    if len(result_tree[pv][ps][component_name][nvr]["sources"]) > 0:
                        source_names = [
                            source["name"]
                            for source in result_tree[pv][ps][component_name][nvr]["sources"]
                            if source.get("namespace") == "REDHAT"
                        ]
                        component_names.update(source_names)
                    else:
                        if ctx.params["no_upstream_affects"]:
                            if result_tree[pv][ps][component_name][nvr]["namespace"] == "REDHAT":
                                component_names.add(component_name)
                        else:
                            component_names.add(component_name)

        # we should only show component name if both {component name} and {component name-container} exists # noqa
        if (
            search_component_name in component_names
            and f"{search_component_name}-container" in component_names
        ):
            component_names.remove(f"{search_component_name}-container")
        if format == "text":
            for cn in component_names:
                # ensure {component name} is not in profile exclude components enum
                if not any([re.search(match, cn) for match in exclude_components]):
                    console.print(
                        f"{pv}/{cn}={flaw_operation}",
                        no_wrap=no_wrap,
                    )
        else:
            affects = []
            for cn in component_names:
                # ensure {component name} is not in profile exclude components enum
                if not any([re.search(match, cn) for match in exclude_components]):
                    affects.append(
                        {"product_version": pv, "component_name": cn, "operation": flaw_operation}
                    )
            return affects


def text_output_products_contain_component(
    ctx,
    output,
    exclude_products,
    exclude_components,
    no_wrap=False,
):
    search_component_name = ctx.params["component_name"]

    # handle single component
    if ctx.params["purl"]:
        ordered_results = sorted(output["results"], key=lambda d: d["ofuri"])
        for item in ordered_results:
            console.print(
                Text(item["ofuri"], style="bold magenta u"),
                no_wrap=no_wrap,
            )
        ctx.exit()

    # handle multiple components
    if "results" in output and output["count"] > 0:
        console.highlighter = None

        # first flatten the tree
        normalised_results = generate_normalised_results(
            output,
            exclude_products,
            exclude_components,
            ctx.params["output_type_filter"],
            ctx.params["include_inactive_product_streams"],
            ctx.params["include_product_stream_excluded_components"],
        )
        result_tree = generate_result_tree(normalised_results)

        # TODO - MAVEN component type will require special handling
        if ctx.params["affect_mode"]:
            console.no_color = True
            console.highlighter = None

            flaw_mode = ctx.params["flaw_mode"]
            flaw_operation = "new"
            if flaw_mode == "add":
                flaw_operation = "new"
            if flaw_mode == "update":
                flaw_operation = "update"
            if flaw_mode == "dry_run":
                flaw_operation = "dry_run"

            generate_affects(ctx, result_tree, exclude_components, flaw_operation, no_wrap=False)

        else:
            if ctx.obj["VERBOSE"] == 0:  # product_version X component_name
                for pv in result_tree.keys():
                    component_names = set()
                    for ps in result_tree[pv].keys():
                        component_names.update(result_tree[pv][ps].keys())
                    # we should only show component name if both {component name} and {component name-container} exists # noqa
                    if (
                        search_component_name in component_names
                        and f"{search_component_name}-container" in component_names
                    ):
                        component_names.remove(f"{search_component_name}-container")
                    for cn in sorted(component_names):
                        # highlight search term
                        dep_name = re.sub(re.escape(cn), f"[b]{cn}[/b]", cn)
                        dep = f"[grey93]{dep_name}[/grey93]"
                        console.print(
                            Text(pv, style="magenta b u"),
                            dep,
                            no_wrap=no_wrap,
                        )
            if ctx.obj["VERBOSE"] == 1:  # product_stream X nvr x related_url
                for pv in result_tree.keys():
                    for ps in result_tree[pv].keys():
                        for cn in sorted(result_tree[pv][ps].keys()):
                            # select the latest nvr (from sorted list)
                            nvr = list(result_tree[pv][ps][cn].keys())[-1]
                            # highlight search term
                            dep_name = nvr
                            try:
                                dep_name = re.sub(
                                    re.escape(search_component_name),
                                    f"[b]{search_component_name}[/b]",
                                    nvr,
                                )
                            except re.error:
                                pass
                            dep = f"[grey93]{dep_name} ({result_tree[pv][ps][cn][nvr]['type']})[/grey93]"  # noqa
                            related_url = result_tree[pv][ps][cn][nvr].get("related_url")
                            try:
                                if result_tree[pv][ps][cn][nvr]["related_url"]:
                                    related_url = re.sub(
                                        re.escape(search_component_name),
                                        f"[b]{search_component_name}[/b]",
                                        result_tree[pv][ps][cn][nvr]["related_url"],
                                    )
                            except re.error:
                                pass
                            if result_tree[pv][ps][cn][nvr]["upstreams"]:
                                upstream_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source["name"]
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "upstreams"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                if len(upstream_component_names) > 0:
                                    upstream_component_name = (
                                        f"[cyan]{upstream_component_names[0]}[/cyan]"
                                    )
                                    if len(upstream_component_names) > 1:
                                        upstream_component_name = f"[cyan]{upstream_component_names[0]} and {len(upstream_component_names) - 1} more[/cyan]"  # noqa
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        upstream_component_name,
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if result_tree[pv][ps][cn][nvr]["sources"]:
                                source_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source["name"]
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "sources"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                if len(source_component_names) > 0:
                                    source_component_name = (
                                        f"[red]{source_component_names[0]}[/red]"
                                    )
                                    if len(source_component_names) > 1:
                                        source_component_name = f"[red]{source_component_names[0]} and {len(source_component_names) - 1} more[/red]"  # noqa
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        source_component_name,
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if not (result_tree[pv][ps][cn][nvr]["upstreams"]) and not (
                                result_tree[pv][ps][cn][nvr]["sources"]
                            ):
                                console.print(
                                    Text(ps, style="magenta b u"),
                                    dep,
                                    f"([grey]{related_url}[/grey])",
                                    no_wrap=no_wrap,
                                )
            if ctx.obj["VERBOSE"] == 2:  # product_stream X nvr x related_url x build_source_url
                for pv in result_tree.keys():
                    for ps in result_tree[pv].keys():
                        for cn in sorted(result_tree[pv][ps].keys()):
                            nvr = list(result_tree[pv][ps][cn].keys())[-1]
                            # highlight search term
                            dep_name = nvr
                            try:
                                dep_name = re.sub(
                                    re.escape(search_component_name),
                                    f"[b]{search_component_name}[/b]",
                                    nvr,
                                )
                            except re.error:
                                pass
                            dep = f"[grey93]{dep_name} ({result_tree[pv][ps][cn][nvr]['type']})[/grey93]"  # noqa
                            related_url = result_tree[pv][ps][cn][nvr].get("related_url")
                            try:
                                if result_tree[pv][ps][cn][nvr]["related_url"]:
                                    related_url = re.sub(
                                        re.escape(search_component_name),
                                        f"[b]{search_component_name}[/b]",
                                        result_tree[pv][ps][cn][nvr]["related_url"],
                                    )
                            except re.error:
                                pass
                            build_source_url = ""
                            if result_tree[pv][ps][cn][nvr]["build_source_url"]:
                                build_source_url = result_tree[pv][ps][cn][nvr]["build_source_url"]
                            if result_tree[pv][ps][cn][nvr]["upstreams"]:
                                upstream_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source.get("name")
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "upstreams"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                if len(upstream_component_names) > 0:
                                    upstream_component_name = (
                                        f"[cyan]{upstream_component_names[0]}[/cyan]"
                                    )
                                    if len(upstream_component_names) > 1:
                                        upstream_component_name = f"[cyan]{upstream_component_names[0]} and {len(upstream_component_names) - 1} more[/cyan]"  # noqa
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        upstream_component_name,
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        f"([grey]{build_source_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if result_tree[pv][ps][cn][nvr]["sources"]:
                                source_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source.get("name")
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "sources"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                if len(source_component_names) > 0:
                                    source_component_name = (
                                        f"[red]{source_component_names[0]}[/red]"
                                    )
                                    if len(source_component_names) > 1:
                                        source_component_name = f"[red]{source_component_names[0]} and {len(source_component_names) - 1} more[/red]"  # noqa
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        source_component_name,
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        f"([grey]{build_source_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if not (result_tree[pv][ps][cn][nvr]["upstreams"]) and not (
                                result_tree[pv][ps][cn][nvr]["sources"]
                            ):
                                console.print(
                                    Text(ps, style="magenta b u"),
                                    dep,
                                    f"([grey]{related_url}[/grey])",
                                    no_wrap=no_wrap,
                                )
            if (
                ctx.obj["VERBOSE"] == 3
            ):  # product_stream X nvr (full source/upstreams) x related_url x build_source_url
                for pv in result_tree.keys():
                    for ps in result_tree[pv].keys():
                        for cn in sorted(result_tree[pv][ps].keys()):
                            # select the latest nvr (from sorted list)
                            nvr = list(result_tree[pv][ps][cn].keys())[-1]
                            # highlight search term
                            dep_name = nvr
                            try:
                                dep_name = re.sub(
                                    re.escape(search_component_name),
                                    f"[b]{search_component_name}[/b]",
                                    nvr,
                                )
                            except re.error:
                                pass
                            dep = f"[grey93]{dep_name} ({result_tree[pv][ps][cn][nvr]['type']})[/grey93]"  # noqa
                            related_url = result_tree[pv][ps][cn][nvr].get("related_url")
                            try:
                                if result_tree[pv][ps][cn][nvr]["related_url"]:
                                    related_url = re.sub(
                                        re.escape(search_component_name),
                                        f"[b]{search_component_name}[/b]",
                                        result_tree[pv][ps][cn][nvr]["related_url"],
                                    )
                            except re.error:
                                pass
                            build_source_url = ""
                            if result_tree[pv][ps][cn][nvr]["build_source_url"]:
                                build_source_url = result_tree[pv][ps][cn][nvr]["build_source_url"]
                            if result_tree[pv][ps][cn][nvr]["upstreams"]:
                                upstream_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source["name"]
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "upstreams"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                for upstream_name in upstream_component_names:
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        f"[cyan]{upstream_name}[/cyan]",
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        f"([grey]{build_source_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if result_tree[pv][ps][cn][nvr]["sources"]:
                                source_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source["name"]
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "sources"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                for source_name in source_component_names:
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        f"[light_blue]{source_name}[/light_blue]",
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        f"([grey]{build_source_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if not (result_tree[pv][ps][cn][nvr]["upstreams"]) and not (
                                result_tree[pv][ps][cn][nvr]["sources"]
                            ):
                                console.print(
                                    Text(ps, style="magenta b u"),
                                    dep,
                                    f"([grey]{related_url}[/grey])",
                                    f"([grey]{build_source_url}[/grey])",
                                    no_wrap=no_wrap,
                                )
            if (
                ctx.obj["VERBOSE"] > 3
            ):  # product_stream X nvr (full source/upstreams) x related_url x build_source_url
                for pv in result_tree.keys():
                    for ps in result_tree[pv].keys():
                        for cn in sorted(result_tree[pv][ps].keys()):
                            # select the latest nvr (from sorted list)
                            nvr = list(result_tree[pv][ps][cn].keys())[-1]
                            # highlight search term
                            dep_name = nvr
                            try:
                                dep_name = re.sub(
                                    re.escape(search_component_name),
                                    f"[b]{search_component_name}[/b]",
                                    nvr,
                                )
                            except re.error:
                                pass
                            dep = f"[grey93]{dep_name} ({result_tree[pv][ps][cn][nvr]['type']})[/grey93]"  # noqa
                            related_url = result_tree[pv][ps][cn][nvr].get("related_url")
                            try:
                                if result_tree[pv][ps][cn][nvr]["related_url"]:
                                    related_url = re.sub(
                                        re.escape(search_component_name),
                                        f"[b]{search_component_name}[/b]",
                                        result_tree[pv][ps][cn][nvr]["related_url"],
                                    )
                            except re.error:
                                pass
                            build_source_url = ""
                            if result_tree[pv][ps][cn][nvr]["build_source_url"]:
                                build_source_url = result_tree[pv][ps][cn][nvr]["build_source_url"]
                            if result_tree[pv][ps][cn][nvr]["upstreams"]:
                                upstream_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source["nvr"]
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "upstreams"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                for upstream_name in upstream_component_names:
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        f"[cyan]{upstream_name}[/cyan]",
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        f"([grey]{build_source_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if result_tree[pv][ps][cn][nvr]["sources"]:
                                source_component_names = sorted(
                                    list(
                                        set(
                                            [
                                                source["nvr"]
                                                for source in result_tree[pv][ps][cn][nvr][
                                                    "sources"
                                                ]
                                            ]
                                        )
                                    )
                                )
                                for source_name in source_component_names:
                                    console.print(
                                        Text(ps, style="magenta b u"),
                                        f"[light_blue]{source_name}[/light_blue]",
                                        dep,
                                        f"([grey]{related_url}[/grey])",
                                        f"([grey]{build_source_url}[/grey])",
                                        no_wrap=no_wrap,
                                    )
                            if not (result_tree[pv][ps][cn][nvr]["upstreams"]) and not (
                                result_tree[pv][ps][cn][nvr]["sources"]
                            ):
                                console.print(
                                    Text(ps, style="magenta b u"),
                                    dep,
                                    f"([grey]{related_url}[/grey])",
                                    f"([grey]{build_source_url}[/grey])",
                                    no_wrap=no_wrap,
                                )

        ctx.exit()


def text_output_components_contain_component(
    ctx, output, format, exclude_components, no_wrap=False
):
    if "results" in output:
        for item in output["results"]:
            component_name = item["name"]
            component_nvr = item["nvr"]
            related_url = item["related_url"]
            download_url = item["download_url"]
            arch = item["arch"]
            if not any([re.search(match, component_name) for match in exclude_components]):
                if ctx.obj["VERBOSE"] == 0:
                    ordered_sources = sorted(item["sources"], key=lambda d: d["purl"])
                    for source in ordered_sources:
                        if "arch=noarch" in source["purl"] or "arch=src" in source["purl"]:
                            source_purl = PackageURL.from_string(source["purl"])
                            root_component = source_purl.name
                            if source_purl.type == "oci" and "-source" not in source_purl.name:
                                root_component = (
                                    f"[u magenta]{source_purl.name}-container[/u magenta]"
                                )
                            if source_purl.type == "oci":
                                component_ns = Text("REDHAT", style="bold magenta")
                            elif not source_purl.namespace:
                                component_ns = Text("UPSTREAM", style="bold magenta")
                            else:
                                component_ns = Text(source_purl.namespace.upper(), style="bold red")

                            console.print(
                                component_ns,
                                source_purl.type.upper(),
                                root_component,
                                Text(component_name, style="bold white"),
                                no_wrap=no_wrap,
                            )
                if ctx.obj["VERBOSE"] == 1:
                    ordered_sources = sorted(item["sources"], key=lambda d: d["purl"])
                    for source in ordered_sources:
                        if "arch=noarch" in source["purl"] or "arch=src" in source["purl"]:
                            source_purl = PackageURL.from_string(source["purl"])
                            root_component = f"{source_purl.name}-{source_purl.version}"
                            if source_purl.type == "oci" and "-source" not in source_purl.name:
                                root_component = (
                                    f"[u magenta]{source_purl.name}-container[/u magenta]"
                                )
                            if source_purl.type == "oci":
                                component_ns = Text("REDHAT", style="bold magenta")
                            elif not source_purl.namespace:
                                component_ns = Text("UPSTREAM", style="bold magenta")
                            else:
                                component_ns = Text(source_purl.namespace.upper(), style="bold red")
                            console.print(
                                component_ns,
                                source_purl.type.upper(),
                                root_component,
                                Text(component_nvr, style="bold white"),
                                arch,
                                no_wrap=no_wrap,
                            )
                if ctx.obj["VERBOSE"] > 1:
                    ordered_sources = sorted(item["sources"], key=lambda d: d["purl"])
                    for source in ordered_sources:
                        if "arch=noarch" in source["purl"] or "arch=src" in source["purl"]:
                            source_purl = PackageURL.from_string(source["purl"])
                            root_component = f"{source_purl.name}-{source_purl.version}"
                            if source_purl.type == "oci" and "-source" not in source_purl.name:
                                root_component = (
                                    f"[u magenta]{source_purl.name}-container[/u magenta]"
                                )
                            if source_purl.type == "oci":
                                component_ns = Text("REDHAT", style="bold magenta")
                            elif not source_purl.namespace:
                                component_ns = Text("UPSTREAM", style="bold magenta")
                            else:
                                component_ns = Text(source_purl.namespace.upper(), style="bold red")
                            console.print(
                                component_ns,
                                source_purl.type.upper(),
                                root_component,
                                Text(component_nvr, style="bold white"),
                                arch,
                                related_url,
                                download_url,
                                no_wrap=no_wrap,
                            )
    ctx.exit()


def text_output_components_affected_by_cve(ctx, output, format, no_wrap=False):
    console.print("Flaw Title:", output["title"])
    console.print(
        "affects:",
    )
    for component in output["components"]:
        affected_component1 = f"({component['purl']})"
        if not ctx.obj["SHOW_PURL"]:
            purl = PackageURL.from_string(component["purl"])
            ns = "UPSTREAM"
            if purl.namespace:
                ns = purl.namespace.upper()
            affected_component = f"([bold cyan]{ns}[/bold cyan] {purl.name}-{purl.version},{purl.type.upper()})"  # noqa
            versions = [pv["name"] for pv in component["product_versions"]]
            if ctx.obj["VERBOSE"] == 0:
                console.print(
                    Text(str(versions), style="bold magenta u"),
                    ns,
                    affected_component,
                    no_wrap=no_wrap,
                )
            if ctx.obj["VERBOSE"] == 1:
                for ps in component["product_streams"]:
                    console.print(
                        Text(ps.get("name"), style="bold magenta u"),
                        ns,
                        affected_component1,
                        no_wrap=no_wrap,
                    )
            if ctx.obj["VERBOSE"] > 1:
                for ps in component["product_streams"]:
                    console.print(
                        Text(ps.get("name"), style="bold magenta u"),
                        ns,
                        affected_component1,
                        Text(component.get("build_source_url", ""), style="i"),
                        Text(component.get("related_url", ""), style="i"),
                        Text(component.get("download_url", ""), style="i"),
                        no_wrap=no_wrap,
                    )
    ctx.exit()


def text_output_products_affected_by_cve(ctx, output, format, exclude_products, no_wrap=False):
    console.print("[white]link:[/white]", output["link"])
    console.print("[white]cve_id:[/white]", output["cve_id"])
    console.print("[white]title:[/white]", output["title"])
    if ctx.obj["VERBOSE"] == 0:
        console.print(
            "[white]product_versions:[/white]",
        )
        ordered_product_versions = sorted(output["product_versions"])
        for product_version in ordered_product_versions:
            console.print(Text(product_version, style="bold magenta u"), no_wrap=no_wrap)
    if ctx.obj["VERBOSE"] > 0:
        console.print(
            "[white]product_streams:[/white]",
        )
        ordered_product_streams = sorted(output["product_streams"])
        for product_stream in ordered_product_streams:
            console.print(Text(product_stream, style="bold magenta u"), no_wrap=True)
    ctx.exit()


def text_output_get_manifest(ctx, output, format, no_wrap=False):
    if not ctx.obj["SHOW_PURL"]:
        for component in output["packages"]:
            if "pkg:" in component["externalRefs"][0]["referenceLocator"]:
                purl = PackageURL.from_string(component["externalRefs"][0]["referenceLocator"])
                ns = "[cyan]UPSTREAM[/cyan]"
                component = f"([bold turquoise2]{ns}[/bold turquoise2] [white]{purl.name}-{purl.version}[/white],{component_type_style(purl.type.upper())})"  # noqa
                if purl.namespace == "redhat":
                    ns = f"[red]{purl.namespace.upper()}[/red]"
                    component = f"([white]{purl.name}-{purl.version}[/white],{component_type_style(purl.type.upper())})"  # noqa
                else:
                    if purl.namespace:
                        ns = f"[white]{purl.namespace.upper()}[/white]"
                    component = f"([white]{purl.name}-{purl.version}[/white],{component_type_style(purl.type.upper())})"  # noqa
                console.print(ns, component, no_wrap=no_wrap)  # noqa
    else:
        for component in output["packages"]:
            purl = component["externalRefs"][0]["referenceLocator"]
            console.print(purl, no_wrap=no_wrap)  # noqa

    ctx.exit()


def text_output_component_flaws(ctx, output, format, no_wrap=False):
    ordered_components = sorted(output["results"], key=lambda d: d["name"])
    for item in ordered_components:
        component_name = item["name"]
        # sorting should work when there is no title or cve-id key
        ordered_affects = sorted(item["affects"], key=lambda d: d["flaw_cve_id"])
        for affect in ordered_affects:
            flaw_cve_id = "Vulnerability"
            if affect["flaw_cve_id"]:
                flaw_cve_id = affect["flaw_cve_id"]
            if ctx.obj["VERBOSE"] == 0:
                console.print(
                    Text(flaw_cve_id, style="magenta"),
                    Text(component_name, style="white"),
                    Text(affect["affect_product_version"], style="cyan"),
                    affect["affect_affectedness"],
                    affect["affect_impact"],
                    affect["affect_resolution"],
                    no_wrap=no_wrap,
                )
            if ctx.obj["VERBOSE"] == 1:
                console.print(
                    Text(flaw_cve_id, style="magenta"),
                    Text(component_name, style="white"),
                    Text(affect["affect_product_version"], style="cyan"),
                    f"(state: {affect['flaw_state']} resolution:{affect['flaw_resolution']})",
                    affect["affect_affectedness"],
                    affect["affect_impact"],
                    affect["affect_resolution"],
                    no_wrap=no_wrap,
                )
            if ctx.obj["VERBOSE"] > 1:
                console.print(Text(affect["title"], style="white"))
                console.print(
                    Text(component_name, style="magenta"),
                    Text(affect["affect_product_version"], style="cyan"),
                    f"(state: {affect['flaw_state']} resolution:{affect['flaw_resolution']})",
                    affect["affect_affectedness"],
                    affect["affect_impact"],
                    affect["affect_resolution"],
                    no_wrap=True,
                )
    ctx.exit()


def text_output_product_flaws(ctx, output, format, no_wrap=False):
    for item in output["results"]:
        component_name = item["name"]
        for affect in item["affects"]:
            flaw_cve_id = "Vulnerability"
            if affect["flaw_cve_id"]:
                flaw_cve_id = affect["flaw_cve_id"]
            if ctx.obj["VERBOSE"] == 0:
                console.print(
                    Text(flaw_cve_id, style="magenta"),
                    Text(component_name, style="white"),
                    affect["affect_component_name"],
                    affect["affect_affectedness"],
                    affect["affect_impact"],
                    affect["affect_resolution"],
                    no_wrap=no_wrap,
                )
            if ctx.obj["VERBOSE"] == 1:
                console.print(
                    Text(flaw_cve_id, style="magenta"),
                    Text(component_name, style="white"),
                    f"(state: {affect['flaw_state']} resolution:{affect['flaw_resolution']})",
                    affect["affect_component_name"],
                    affect["affect_affectedness"],
                    affect["affect_impact"],
                    affect["affect_resolution"],
                    no_wrap=no_wrap,
                )
            if ctx.obj["VERBOSE"] > 1:
                console.print(affect["title"])
                console.print(
                    Text(flaw_cve_id, style="magenta"),
                    Text(component_name, style="white"),
                    f"(state: {affect['flaw_state']} resolution:{affect['flaw_resolution']})",
                    affect["affect_component_name"],
                    affect["affect_affectedness"],
                    affect["affect_impact"],
                    affect["affect_resolution"],
                    no_wrap=no_wrap,
                )
    ctx.exit()


def text_output_list(ctx, output, format, exclude_components, no_wrap=False):
    if "results" in output and output["count"] > 0:
        # handle component
        if "purl" in output["results"][0]:
            ordered_components = sorted(output["results"], key=lambda d: d["name"])
            for row in ordered_components:
                if not any([re.search(match, row["purl"]) for match in exclude_components]):
                    if "purl" in row:
                        purl = PackageURL.from_string(row["purl"])
                        if purl.type == "oci":
                            component_ns = Text("REDHAT", style="bold magenta")
                        elif not purl.namespace:
                            component_ns = Text("UPSTREAM", style="bold magenta")
                        else:
                            component_ns = Text(purl.namespace.upper(), style="bold red")

                        sha256 = ""
                        if purl.version:
                            if purl.version.startswith("sha256"):
                                sha256 = output_version(ctx, purl.version)
                        nvr = None
                        if "nvr" in row:
                            nvr = row["nvr"]

                        if not ctx.obj["SHOW_PURL"]:
                            if ctx.obj["VERBOSE"] == 0:
                                console.print(
                                    component_ns,
                                    purl.type.upper(),
                                    Text(nvr, style="bold white"),
                                    sha256,
                                    row["related_url"],
                                    purl.qualifiers.get("arch"),
                                )
                            if ctx.obj["VERBOSE"] > 0:
                                download_url = ""
                                if "download_url" in row:
                                    download_url = row["download_url"]
                                console.print(
                                    component_ns,
                                    purl.type.upper(),
                                    Text(nvr, style="bold white"),
                                    sha256,
                                    row["related_url"],
                                    purl.qualifiers.get("arch"),
                                    download_url,
                                )
                        else:
                            console.print(
                                row["purl"],
                            )
        # handle flaw
        if "cve_id" in output["results"][0]:
            for row in output["results"]:
                console.print(
                    row["title"],
                    row["state"],
                    row["impact"],
                    row["resolution"],
                    no_wrap=no_wrap,
                )

        # handle trackers
        if all(key in output["results"][0] for key in ("external_system_id", "status")):
            for row in output["results"]:
                console.print(
                    row["external_system_id"],
                    row["type"],
                    row["status"],
                    no_wrap=no_wrap,
                )

        # handle products
        if "ofuri" in output["results"][0]:
            for row in output["results"]:
                console.print(
                    Text(row["name"], style="magenta bold u"),
                    row["ofuri"],
                    no_wrap=no_wrap,
                )
        # handle channels
        if "relative_url" in output["results"][0]:
            for row in output["results"]:
                console.print(
                    Text(row["name"], style="magenta bold u"),
                    row["type"],
                    row["description"],
                    no_wrap=no_wrap,
                )
        # handle builds
        if "build_id" in output["results"][0]:
            for row in output["results"]:
                console.print(
                    Text(str(row["build_id"]), style="magenta bold u"),
                    row["build_type"],
                    Text(row["name"], style="white"),
                    row["created_at"],
                    Text(row["link"], style="i"),
                    no_wrap=no_wrap,
                )

    ctx.exit()


def text_output_purls(ctx, output, format, no_wrap=False):
    if "results" in output and output["count"] > 0:
        # handle component
        if "purl" in output["results"][0]:
            ordered_components = sorted(output["results"], key=lambda d: d["purl"])
            for row in ordered_components:
                if "purl" in row:
                    purl = PackageURL.from_string(row["purl"])
                    if not purl.namespace:
                        component_ns = Text("UPSTREAM", style="bold magenta")
                    else:
                        component_ns = Text(purl.namespace.upper(), style="bold red")
                    if ctx.obj["VERBOSE"] == 0:
                        console.print(
                            component_ns,
                            purl.type.upper(),
                            Text(purl.name, style="bold white"),
                            purl.version,
                            purl.qualifiers.get("arch"),
                            no_wrap=no_wrap,
                        )
                    if ctx.obj["VERBOSE"] > 0:
                        console.print(
                            component_ns,
                            purl.type.upper(),
                            Text(purl.name, style="bold white"),
                            purl.version,
                            purl.qualifiers.get("arch"),
                            row["link"],
                            no_wrap=no_wrap,
                        )
        ctx.exit()


def text_output_generic(ctx, output, format, no_wrap=False):
    for k, v in output.items():
        key_name = Text(k)
        key_name.stylize("bold magenta")
        console.print(key_name, " : ", v, no_wrap=no_wrap)


def cprint(
    data,
    dest=DEST.CONSOLE,
    filename=None,
    ctx=None,
    show_count: bool = True,
):
    """handle format and output"""
    from griffon import get_config_option

    exclude_products = []
    if get_config_option(ctx.obj["PROFILE"], "exclude"):
        exclude_products = get_config_option(ctx.obj["PROFILE"], "exclude").split("\n")
    logger.debug(f"exclude products = {exclude_products}")

    exclude_components = []
    if get_config_option(ctx.obj["PROFILE"], "exclude_components"):
        exclude_components = get_config_option(ctx.obj["PROFILE"], "exclude_components").split("\n")
    logger.debug(f"exclude_components = {exclude_components}")

    output = raw_json_transform(data, show_count)
    if ctx and ctx.obj["NO_COLOR"]:
        console.no_color = True
    format = OUTPUT_FORMAT.JSON
    if ctx and "FORMAT" in ctx.obj:
        format = OUTPUT_FORMAT(ctx.obj["FORMAT"])
    if format is OUTPUT_FORMAT.TEXT:
        no_wrap = ctx.obj["NO_WRAP"]
        terminal_width = ctx.obj["TERMINAL_WIDTH"]
        console.width = int(terminal_width)
        if ctx.info_name == "product-summary":
            text_output_product_summary(ctx, output, format, exclude_products, no_wrap=no_wrap)
        if ctx.info_name == "products-contain-component":
            text_output_products_contain_component(
                ctx,
                output,
                exclude_products,
                exclude_components,
                no_wrap=no_wrap,
            )
        if ctx.info_name == "components-contain-component":
            text_output_components_contain_component(
                ctx, output, format, exclude_components, no_wrap=no_wrap
            )
        if ctx.info_name == "components-affected-by-flaw":
            text_output_components_affected_by_cve(ctx, output, format, no_wrap=no_wrap)
        if ctx.info_name == "products-affected-by-flaw":
            text_output_products_affected_by_cve(
                ctx, output, format, exclude_products, no_wrap=no_wrap
            )
        if ctx.info_name == "get-manifest":
            text_output_get_manifest(ctx, output, format, no_wrap=no_wrap)
        if ctx.info_name == "list":
            text_output_list(ctx, output, format, exclude_components, no_wrap=no_wrap)
        if ctx.info_name == "component-flaws":
            text_output_component_flaws(ctx, output, format, no_wrap=no_wrap)
        if ctx.info_name == "product-flaws":
            text_output_product_flaws(ctx, output, format, no_wrap=no_wrap)
        if ctx.info_name == "provides":
            text_output_purls(ctx, output, format, no_wrap=no_wrap)
        if ctx.info_name == "sources":
            text_output_purls(ctx, output, format, no_wrap=no_wrap)
        if ctx.info_name == "tree":
            text_output_tree(ctx, output, no_wrap=no_wrap)

        # last chance text formatted output
        text_output_generic(ctx, output, format, no_wrap=no_wrap)

    if format is OUTPUT_FORMAT.JSON:
        if dest is DEST.CONSOLE:
            console.print_json(json.dumps(output))

    # if we instructed to open browser, open that up now
    if ctx:
        if "link" in data and "open_browser" in ctx.obj:
            if ctx.obj["open_browser"]:
                click.launch(data["link"])

    exit(0)



"""
    report cli commands

"""
import logging

import click

from griffon import progress_bar
from griffon.autocomplete import (
    get_component_names,
    get_component_purls,
    get_product_stream_names,
    get_product_stream_ofuris,
)
from griffon.output import cprint
from griffon.services import ReportService, core_reports

logger = logging.getLogger("griffon")

report_service = ReportService()


@click.group(name="reports", help="Generate reports.")
@click.pass_context
def reports_grp(ctx):
    """Report operations."""
    pass


@reports_grp.command(name="report-affects", help="Generate Affects example report.")
@click.argument("product_version_name", required=False)
@click.option(
    "--show-components", is_flag=True, default=False, help="Show specific component counts."
)
@click.option("--all", is_flag=True, default=False, help="Show summary report on all affects.")
@click.option("--show-products", is_flag=True, default=False, help="Show specific product counts.")
@click.option("--purl", shell_complete=get_component_purls)
@click.option("--name", shell_complete=get_component_names)
@click.option("--product-name", shell_complete=get_product_stream_names)
@click.option("--ofuri", shell_complete=get_product_stream_ofuris)
@click.pass_context
@progress_bar
def generate_affects_report(
    ctx, product_version_name, all, show_components, show_products, purl, name, product_name, ofuri
):
    """A report operation"""
    if not all and not product_version_name:
        click.echo(ctx.get_help())
        exit(0)
    cprint(report_service.invoke(core_reports.example_affects_report, ctx.params), ctx=ctx)


@reports_grp.command(name="report-entities", help="Generate Entity report (with counts).")
@click.option("--all", is_flag=True, default=True, help="Show summary report on all entities.")
@click.pass_context
@progress_bar
def generate_entity_report(ctx, all):
    """A report operation"""
    if not all:
        click.echo(ctx.get_help())
        exit(0)
    cprint(report_service.invoke(core_reports.entity_report, ctx.params), ctx=ctx)


@reports_grp.command(name="report-license", help="Generate Product Stream license report.")
@click.argument(
    "product_stream_name",
    required=False,
    type=click.STRING,
    shell_complete=get_product_stream_names,
)
@click.option("--purl", help="Component Purl (must be quoted).")
@click.option(
    "--exclude_children", is_flag=True, default=False, help="Exclude children Component licenses."
)
@click.pass_context
@progress_bar
def generate_license_report(ctx, product_stream_name, purl, exclude_children):
    """A report operation"""
    if not product_stream_name and not purl:
        click.echo(ctx.get_help())
        exit(0)
    ctx.obj["FORMAT"] = "json"
    cprint(report_service.invoke(core_reports.license_report, ctx.params), ctx=ctx)



"""
osv plugin:

https://osv.dev/docs/#tag/api/operation/OSV_QueryAffected

"""
import json
import logging

import click
import requests

from griffon.output import cprint

logger = logging.getLogger("griffon")

api_url = "https://api.osv.dev/v1/query"
search_api_url = "https://osv.dev/list"


@click.group()
@click.pass_context
def plugins(ctx):
    """OSV plugin"""
    pass


@plugins.command()
@click.option("--version", "package_version", help="package version")
@click.option("--name", "package_name", help="package name")
@click.option("--ecosystem", help="ecosystem (ex. PyPI) ")
@click.pass_context
def query_by_version(ctx, package_version, package_name, ecosystem):
    if not package_version and not package_name:
        click.echo(ctx.get_help())
        exit(0)
    data = json.dumps(
        {"version": package_version, "package": {"name": package_name, "ecosystem": ecosystem}}
    )
    res = requests.post(
        api_url,
        data=data,
        headers={"Content-type": "application/json"},
    )
    cprint(res.json(), ctx=ctx)


@plugins.command()
@click.option("--commit_hash", help="specific commit hash")
@click.pass_context
def query_by_commit_hash(ctx, commit_hash):
    if not commit_hash:
        click.echo(ctx.get_help())
        exit(0)
    data = json.dumps({"commit": commit_hash})
    res = requests.post(
        api_url,
        data=data,
        headers={"Content-type": "application/json"},
    )
    cprint(res.json(), ctx=ctx)


@plugins.command()
@click.argument(
    "query_string",
    required=True,
    type=click.STRING,
)
@click.pass_context
def search(ctx, query_string):
    """Search osv.dev by query term"""
    if not query_string:
        click.echo(ctx.get_help())
        exit(0)
    click.launch(f"{search_api_url}?ecosystem=&q={query_string}")



"""
    mutation cli commands

"""
import logging

import click
from requests import HTTPError

from griffon import progress_bar
from griffon.exceptions import catch_exception
from griffon.output import cprint
from griffon.services import ProcessService, core_process

logger = logging.getLogger("griffon")

process_service = ProcessService()


# @click.group(name="process", help="Service operations that perform mutations/write.")
# @click.pass_context
# def process_grp(ctx):
#     """Mutation operations."""
#     pass


@click.command(
    name="generate_affects_for_component", help="(UNDER DEV)Generate affects for component."
)
@click.option("--purl")
@click.option("--cve_id")
@catch_exception(handle=(HTTPError))
@click.pass_context
@progress_bar
def generate_affects_for_component_process(ctx, purl, cve_id):
    """Generate affects for specific component."""
    if not purl and not cve_id:
        click.echo(ctx.get_help())
        exit(0)
    cprint(
        process_service.invoke(
            core_process.generate_affects_for_specific_component_process, ctx.params
        ),
        ctx=ctx,
    )



"""
    service library designed for reuse

"""
# define interface for query which is asserted by mypy as well as runtime checking
import logging
import typing
from typing import Any, Dict, List, Protocol, runtime_checkable

logger = logging.getLogger("griffon")


@runtime_checkable
class Query(Protocol):
    """Query service interface"""

    name: str
    description: str
    allowed_params: list

    def __init__(self, params: dict) -> None:
        """Use generic params dict to pass in all parameters"""
        pass

    def execute(self) -> typing.Union[dict, List[Dict[str, Any]]]:
        """execute() uses a generic ctx dict to pass in all parameters"""
        return {}


@runtime_checkable
class Process(Protocol):
    """Process service interface"""

    name: str
    description: str
    allowed_params: list

    def __init__(self, params: dict) -> None:
        """Use generic params dict to pass in all parameters"""
        pass

    def update(self) -> typing.Union[dict, List[Dict[str, Any]]]:
        return {}


@runtime_checkable
class Report(Protocol):
    """Query service interface"""

    name: str
    description: str
    allowed_params: list

    def __init__(self, params: dict) -> None:
        """Use generic params dict to pass in all parameters"""
        pass

    def generate(self) -> typing.Union[dict, List[Dict[str, Any]]]:
        """execute() uses a generic ctx dict to pass in all parameters"""
        return {}


def check_allowed_params(allowed_params, params):
    """Check if params are allowed."""
    logger.debug(params)
    if allowed_params:
        for key in params.keys():
            assert key in allowed_params


class QueryService:
    def invoke(self, obj, params: dict, status=None):
        check_allowed_params(obj.allowed_params, params)
        return obj(params).execute(status=status)


class ReportService:
    def invoke(self, obj, params: dict, status=None):
        check_allowed_params(obj.allowed_params, params)
        return obj(params).generate()


class ProcessService:
    def invoke(self, obj, params: dict, status=None):
        check_allowed_params(obj.allowed_params, params)
        return obj(params).process()



"""
    read only cli commands

    Note - the command op hierarchy is verbose by design as we will want to reuse
    these operations beyond cli

"""
import copy
import logging
import subprocess
from json import loads

import click
from component_registry_bindings.bindings.python_client.api.v1 import v1_components_list
from component_registry_bindings.bindings.python_client.models import Component

from griffon import (
    MIDDLEWARE_CLI,
    CorgiService,
    OSIDBService,
    get_config_option,
    progress_bar,
)
from griffon.autocomplete import (
    get_component_names,
    get_cve_ids,
    get_product_stream_names,
    get_product_stream_ofuris,
    get_product_version_names,
)
from griffon.commands.entities.corgi import (
    get_component_manifest,
    get_component_summary,
    get_product_stream_manifest,
    list_components,
)
from griffon.commands.entities.helpers import query_params_options
from griffon.commands.reports import (
    generate_affects_report,
    generate_entity_report,
    generate_license_report,
)
from griffon.output import (
    console,
    cprint,
    generate_affects,
    generate_normalised_results,
    generate_result_tree,
    raw_json_transform,
)
from griffon.services import QueryService, core_queries  # , exp

logger = logging.getLogger("griffon")

query_service = QueryService()


@click.group(name="service", help="Service operations.")
@click.pass_context
def queries_grp(ctx):
    """Query operations."""
    pass


queries_grp.add_command(generate_affects_report)
queries_grp.add_command(generate_entity_report)
queries_grp.add_command(generate_license_report)

# queries_grp.add_command(generate_affects_for_component_process)


@queries_grp.command(
    name="product-summary",
    help="Get Product summaries.",
)
@click.argument(
    "product_stream_name",
    required=False,
    type=click.STRING,
    shell_complete=get_product_stream_names,
)
# TODO - underlying bindings need to support 'ofuri'
# @click.option(
#     "--ofuri",
#     "ofuri",
#     type=click.STRING,
#     shell_complete=get_product_stream_ofuris,
#     help="UNDER DEVELOPMENT",
# )
@click.option(
    "-s",
    "strict_name_search",
    is_flag=True,
    default=False,
    help="Strict search, exact match of name.",
)
@click.option(
    "--all",
    "all",
    is_flag=True,
    default=False,
    help="Return all Products.",
)
@click.option(
    "-v",
    "verbose",
    count=True,
    default=get_config_option("default", "verbosity", 0),
    help="Verbose output, more detailed search results, can be used multiple times (e.g. -vvv).",
)  # noqa
@click.pass_context
@progress_bar
def get_product_summary(ctx, product_stream_name, strict_name_search, all, verbose):
    """get product stream."""
    if verbose:
        ctx.obj["VERBOSE"] = verbose
    if not product_stream_name and not all and not strict_name_search:
        click.echo(ctx.get_help())
        exit(0)
    q = query_service.invoke(core_queries.product_stream_summary, ctx.params)
    cprint(q, ctx=ctx)


@queries_grp.command(
    name="component-summary",
    help="Get Component summaries.",
)
@click.argument(
    "component_name",
    required=False,
    type=click.STRING,
    shell_complete=get_component_names,
)
@click.option(
    "-s",
    "strict_name_search",
    is_flag=True,
    default=False,
    help="Strict search, exact match of name.",
)
@click.pass_context
def retrieve_component_summary(ctx, component_name, strict_name_search):
    """Get Component summary."""
    if not component_name:
        click.echo(ctx.get_help())
        exit(0)
    cond = {}
    if component_name:
        cond["component_name"] = component_name
    ctx.invoke(get_component_summary, **cond)


@queries_grp.command(
    name="products-contain-component",
    help="List Products containing Component.",
)
@click.argument(
    "component_name",
    required=False,
)
@click.option(
    "--purl", help="Component purl, needs to be in quotes (ex. 'pkg:rpm/python-pyjwt@1.7.1')"
)
@click.option(
    "--arch",
    default="src",
    type=click.Choice(CorgiService.get_component_arches()),
)
@click.option(
    "--namespace", default=None, type=click.Choice(CorgiService.get_component_namespaces())
)
@click.option("--type", "component_type", type=click.Choice(CorgiService.get_component_types()))
@click.option(
    "-s",
    "strict_name_search",
    is_flag=True,
    default=False,
    help="Strict search, exact match of component name.",
)
@click.option(
    "--generate-affects",
    "-a",
    "affect_mode",
    is_flag=True,
    default=False,
    help="Generate Affects output.",
)
# @click.option(
#     "--cve-id",
#     "cve_id",
#     help="Attach affects to this (\033[1mCVE-ID\033[0m).",
# )
@click.option(
    "--sfm2-flaw-id",
    "sfm2_flaw_id",
    help="Attach affects to this (\033[1msfm2 flaw id\033[0m).",
)
@click.option(
    "--flaw-mode",
    "flaw_mode",
    default="add",
    type=click.Choice(["add", "replace", "dry_run"]),
    help="Add or update when generating affects.",
)
@click.option(
    "--search-latest",
    "search_latest",
    is_flag=True,
    default=False,
    help="Search root Components (\033[1menabled by default\033[0m).",
)
@click.option(
    "--search-related-url",
    "search_related_url",
    is_flag=True,
    default=False,
    help="Search related url (\033[1menabled by default\033[0m).",
)
@click.option(
    "--filter-rh-naming",
    is_flag=True,
    default=False,
    help="Filter rh naming.",
)
@click.option(
    "--search-all",
    "search_all",
    is_flag=True,
    default=False,
    help="Search all Components and dependencies.",
)
@click.option(
    "--search-all-roots",
    "search_all_roots",
    is_flag=True,
    default=False,
    help="Search all ROOT Components and dependencies.",
)
@click.option(
    "--search-redhat",
    "search_redhat",
    is_flag=True,
    default=False,
    help="Search all Red Hat root Components and dependencies.",
)
@click.option(
    "--search-community",
    "search_community",
    is_flag=True,
    default=False,
    help="Search community Components.",
)
@click.option(
    "--search-upstreams",
    "search_upstreams",
    is_flag=True,
    default=False,
    help="Search for Components by upstream.",
)
@click.option(
    "--no-community",
    "no_community",
    is_flag=True,
    default=False,
    help="Do not search community.",
)
@click.option(
    "--no-middleware",
    "no_middleware",
    is_flag=True,
    default=False,
    help="Do not search middleware.",
)
@click.option(
    "--no-upstream-affects",
    "no_upstream_affects",
    is_flag=True,
    default=False,
    help="Do not generate upstream affects.",
)
@click.option(
    "--include-inactive-product-streams",
    "include_inactive_product_streams",
    is_flag=True,
    default=get_config_option("default", "include_inactive_product_streams", False),
    help="Include components from inactive product streams.",
)
@click.option(
    "--include-product-streams-excluded-components",
    "include_product_stream_excluded_components",
    is_flag=True,
    default=get_config_option("default", "include_product_stream_excluded_components", False),
    help="Include product stream excluded components.",
)
@click.option(
    "--output-type-filter",
    "output_type_filter",
    type=click.Choice(CorgiService.get_component_types()),
    default=None,
    help="Filter component type.",
)
@click.option(
    "-v",
    "verbose",
    count=True,
    default=get_config_option("default", "verbosity", 0),
    help="Verbose output, more detailed search results, can be used multiple times (e.g. -vvv).",
)  # noqa
@click.pass_context
def get_product_contain_component(
    ctx,
    component_name,
    purl,
    arch,
    namespace,
    component_type,
    strict_name_search,
    affect_mode,
    sfm2_flaw_id,
    flaw_mode,
    search_latest,
    search_related_url,
    filter_rh_naming,
    search_all,
    search_all_roots,
    search_redhat,
    search_community,
    search_upstreams,
    no_community,
    no_middleware,
    no_upstream_affects,
    include_inactive_product_streams,
    include_product_stream_excluded_components,
    output_type_filter,
    verbose,
):
    with console.status("griffoning", spinner="line") as operation_status:
        """List products of a latest component."""
        if verbose:
            ctx.obj["VERBOSE"] = verbose
        if not purl and not component_name:
            click.echo(ctx.get_help())
            click.echo("")
            click.echo("\033[1mMust supply Component name or --purl.\033[0m")
            exit(0)

        if (
            not search_latest
            and not search_all
            and not search_all_roots
            and not search_related_url
            and not search_community
            and not search_upstreams
            and not search_redhat
        ):
            ctx.params["search_latest"] = True

        params = copy.deepcopy(ctx.params)
        params.pop("verbose")
        params.pop("sfm2_flaw_id")
        params.pop("flaw_mode")
        params.pop("affect_mode")
        if component_name:
            q = query_service.invoke(
                core_queries.products_containing_component_query, params, status=operation_status
            )
        if purl:
            q = query_service.invoke(
                core_queries.products_containing_specific_component_query,
                params,
                status=operation_status,
            )

        # TODO: interim hack for middleware
        if component_name and MIDDLEWARE_CLI and not no_middleware:
            operation_status.update("griffoning: searching deptopia.", spinner="line")
            mw_command = [MIDDLEWARE_CLI, component_name, "-e", "maven", "--json"]
            if strict_name_search:
                mw_command.append("-s")
            proc = subprocess.run(
                mw_command,
                capture_output=True,
                text=True,
            )
            mw_json = loads(proc.stdout)
            mw_components = mw_json["deps"]
            # TODO: need to determine if we use "build" or "deps"
            # if search_all:
            #     mw_components.extend(mw_json["deps"])
            for build in mw_components:
                if build["build_type"] == "maven":
                    component = {
                        "product_versions": [{"name": build["ps_module"]}],
                        "product_streams": [
                            {
                                "name": build["ps_update_stream"],
                                "product_versions": [{"name": build["ps_module"]}],
                            }
                        ],
                        "product_active": True,
                        "type": build["build_type"],
                        "name": build["build_name"],
                        "nvr": build["build_nvr"],
                        "upstreams": [],
                        "sources": [],
                        "software_build": {
                            "build_id": build["build_id"],
                            "source": build["build_repo"],
                        },
                    }
                    if "sources" in build:
                        for deps in build["sources"]:
                            for dep in deps["dependencies"]:
                                components = []
                                components.append(
                                    {
                                        "name": dep.get("name"),
                                        "nvr": dep.get("nvr"),
                                        "type": dep.get("type"),
                                    }
                                )
                                component["sources"] = components
                    q.append(component)

        # TODO: in the short term affect handling will be mediated via sfm2 here in the operation itself # noqa
        if ctx.params["sfm2_flaw_id"]:
            operation_status.update("griffoning: invoking sfm2.", spinner="line")

            console.no_color = True
            console.highlighter = None
            operation_status.stop()

            # generate affects
            output = raw_json_transform(q, True)

            exclude_products = []
            if get_config_option(ctx.obj["PROFILE"], "exclude"):
                exclude_products = get_config_option(ctx.obj["PROFILE"], "exclude").split("\n")
            exclude_components = []
            if get_config_option(ctx.obj["PROFILE"], "exclude_components"):
                exclude_components = get_config_option(
                    ctx.obj["PROFILE"], "exclude_components"
                ).split("\n")
            normalised_results = generate_normalised_results(
                output,
                exclude_products,
                exclude_components,
                output_type_filter,
                include_inactive_product_streams,
                include_product_stream_excluded_components,
            )
            result_tree = generate_result_tree(normalised_results)
            affects = generate_affects(ctx, result_tree, exclude_components, "add", format="json")

            # attempt to import sfm2client module
            try:
                import sfm2client
            except ImportError:
                logger.warning("sfm2client library not found, cannot compare with flaw affects")
                ctx.exit()

            # TODO: paramaterise into dotfile/env var
            sfm2 = sfm2client.api.core.SFMApi({"url": "http://localhost:5600"})
            try:
                flaw = sfm2.flaw.get(sfm2_flaw_id)
            except Exception as e:
                logger.warning(f"Could not retrieve flaw {sfm2_flaw_id}: {e}")
                return

            if ctx.params["flaw_mode"] == "replace":
                if affects:
                    console.print(
                        f"The following affects will REPLACE all flaw {flaw['id']}'s existing affects in \"new\" state:\n"  # noqa
                    )
                    for m in affects:
                        console.print(
                            f"{m['product_version']}\t{m['component_name']}",
                            no_wrap=False,
                        )

                    if click.confirm(
                        f"\nREPLACE flaw {flaw['id']}'s existing affects in \"new\" state with the above? THIS CANNOT BE UNDONE: ",  # noqa
                        default=True,
                    ):
                        click.echo("Updating ...")
                        # only discard affects in 'new' state, we should preserve all others so not to throw work away # noqa
                        new_affects = [a for a in flaw["affects"] if a["affected"] != "new"]
                        # get map of existing affects first, so that we don't try to add duplicates
                        existing = set((a["ps_module"], a["ps_component"]) for a in new_affects)
                        for m in affects:
                            if (m["product_version"], m["component_name"]) in existing:
                                continue
                            new_affects.append(
                                {
                                    "affected": "new",
                                    "ps_component": m["component_name"],
                                    "ps_module": m["product_version"],
                                }
                            )
                        try:
                            sfm2.flaw.update(flaw["id"], data={"affects": new_affects})
                        except Exception as e:
                            msg = e.response.json()
                            logger.warning(f"Failed to update flaw: {e}: {msg}")
                        console.print("Operation done.")
                        ctx.exit()

                    click.echo("No affects were added to flaw.")
                else:
                    console.print("No affects to add to flaw.")
            else:
                missing = []
                for affect in affects:
                    flaw_has_affect = False
                    for a in flaw.get("affects"):
                        if a.get("ps_module") == affect.get("product_version") and a.get(
                            "ps_component"
                        ) == affect.get("component_name"):
                            flaw_has_affect = True
                    if not flaw_has_affect:
                        missing.append(affect)

                if missing:
                    console.log("Flaw is missing the following affect entries:\n")
                    for m in missing:
                        console.print(
                            f"{m['product_version']}\t{m['component_name']}",
                            no_wrap=False,
                        )
                    if click.confirm(
                        "Would you like to add them? ",
                        default=True,
                    ):
                        click.echo("Updating ...")

                        updated_affects = flaw.get("affects")[:]
                        for m in missing:
                            updated_affects.append(
                                {
                                    "affected": "new",
                                    "ps_component": m["component_name"],
                                    "ps_module": m["product_version"],
                                }
                            )
                        try:
                            sfm2.flaw.update(flaw["id"], data={"affects": updated_affects})
                        except Exception as e:
                            msg = e.response.json()
                            logger.warning(f"Failed to update flaw: {e}: {msg}")
                        console.print("Operation done.")
                        ctx.exit()
                    click.echo("No affects were added to flaw.")

                else:
                    console.print("Flaw is not missing any affect entries")

            ctx.exit()

        cprint(q, ctx=ctx)


@queries_grp.command(
    name="components-contain-component",
    help="List Components containing Component.",
)
@click.argument("component_name", required=False)
@click.option("--purl")
@click.option("--type", "component_type", type=click.Choice(CorgiService.get_component_types()))
@click.option("--version", "component_version")
@click.option(
    "--arch",
    "component_arch",
    type=click.Choice(CorgiService.get_component_arches()),
    help="Default arch=src.",
)
@click.option("--namespace", type=click.Choice(CorgiService.get_component_namespaces()))
@click.option(
    "-s",
    "strict_name_search",
    is_flag=True,
    default=False,
    help="Strict search, exact match of component name.",
)
@click.option(
    "-v",
    "verbose",
    count=True,
    default=get_config_option("default", "verbosity", 0),
    help="Verbose output, more detailed search results, can be used multiple times (e.g. -vvv).",
)  # noqa
@click.pass_context
@progress_bar
def get_component_contain_component(
    ctx,
    component_name,
    purl,
    component_type,
    component_version,
    component_arch,
    namespace,
    strict_name_search,
    verbose,
):
    """List components that contain component."""
    if verbose:
        ctx.obj["VERBOSE"] = verbose
    if not component_name and not purl:
        click.echo(ctx.get_help())
        exit(0)
    if component_name:
        q = query_service.invoke(core_queries.components_containing_component_query, ctx.params)
        cprint(q, ctx=ctx)
    if purl:
        q = query_service.invoke(
            core_queries.components_containing_specific_component_query, ctx.params
        )
        cprint(q, ctx=ctx)


@queries_grp.command(
    name="product-manifest",
    help="Get Product manifest (includes Root Components and all dependencies).",
)
@click.argument("product_stream_name", required=False, shell_complete=get_product_stream_names)
@click.option("--ofuri", "ofuri", type=click.STRING, shell_complete=get_product_stream_ofuris)
@click.option(
    "--spdx-json",
    "spdx_json_format",
    is_flag=True,
    default=False,
    help="Generate spdx manifest (json).",
)
@click.pass_context
def get_product_manifest_query(ctx, product_stream_name, ofuri, spdx_json_format):
    """List components of a specific product version."""
    if not ofuri and not product_stream_name:
        click.echo(ctx.get_help())
        exit(0)

    if spdx_json_format:
        ctx.ensure_object(dict)
        ctx.obj["FORMAT"] = "json"  # TODO - investigate if we need yaml format.

    cond = {}
    if ofuri:
        cond["ofuri"] = ofuri
    if product_stream_name:
        cond["product_stream_name"] = product_stream_name
    ctx.invoke(get_product_stream_manifest, **cond)


@queries_grp.command(
    name="product-components",
    help="List LATEST Root Components of Product.",
)
@click.argument("product_stream_name", required=False, shell_complete=get_product_stream_names)
@click.option("--ofuri", "ofuri", type=click.STRING, shell_complete=get_product_stream_ofuris)
@query_params_options(
    entity="Component",
    endpoint_module=v1_components_list,
    options_overrides={
        "include_fields": {"type": click.Choice(CorgiService.get_fields(Component))},
    },
)
@click.option(
    "-v",
    "verbose",
    count=True,
    default=get_config_option("default", "verbosity", 0),
    help="Verbose output, more detailed search results, can be used multiple times (e.g. -vvv).",
)  # noqa
@click.pass_context
def get_product_latest_components_query(ctx, product_stream_name, ofuri, verbose, **params):
    """List components of a specific product version."""
    if verbose:
        ctx.obj["VERBOSE"] = verbose
        ctx.params.pop("verbose")
    if not ofuri and not product_stream_name:
        click.echo(ctx.get_help())
        exit(0)
    if ofuri:
        params["ofuri"] = ofuri
    if product_stream_name:
        # lookup ofuri
        session = CorgiService.create_session()
        ps = session.product_streams.retrieve_list(
            name=product_stream_name, include_fields="name,ofuri"
        )
        params["ofuri"] = ps["ofuri"]
        params["include_fields"] = "name,nvr,related_url,purl,version,release,type,software_build"
    ctx.invoke(list_components, **params)


@queries_grp.command(
    name="component-manifest",
    help="Get Component manifest.",
)
@click.option("--uuid", "component_uuid")
@click.option("--purl", help="Component Purl (must be quoted).")
@click.option(
    "--spdx-json",
    "spdx_json_format",
    is_flag=True,
    default=False,
    help="Generate spdx manifest (json).",
)
@click.pass_context
def retrieve_component_manifest(ctx, component_uuid, purl, spdx_json_format):
    """Retrieve Component manifest."""
    if not component_uuid and not purl:
        click.echo(ctx.get_help())
        exit(0)
    if spdx_json_format:
        ctx.ensure_object(dict)
        ctx.obj["FORMAT"] = "json"
    cond = {}
    if component_uuid:
        cond["uuid"] = component_uuid
    if purl:
        cond["purl"] = purl
    ctx.invoke(get_component_manifest, **cond)


@queries_grp.command(
    name="components-affected-by-flaw",
    help="List Components affected by Flaw.",
)
@click.argument("cve_id", required=False, type=click.STRING, shell_complete=get_cve_ids)
@click.option(
    "--affectedness",
    help="Filter by Affect affectedness.",
    type=click.Choice(OSIDBService.get_affect_affectedness()),
)
@click.option(
    "--resolution",
    "affect_resolution",
    help="Filter by Affect resolution.",
    type=click.Choice(OSIDBService.get_affect_resolution()),
)
@click.option(
    "--impact",
    "affect_impact",
    help="Filter by Affect impact.",
    type=click.Choice(OSIDBService.get_affect_impact()),
)
@click.option(
    "--type",
    "component_type",
    type=click.Choice(CorgiService.get_component_types()),
    help="Filter by Component type.",
)
@click.option(
    "--namespace",
    type=click.Choice(CorgiService.get_component_namespaces()),
    help="filter by Component namespace.",
)
@click.pass_context
@progress_bar
def components_affected_by_specific_cve_query(
    ctx, cve_id, affectedness, affect_resolution, affect_impact, component_type, namespace
):
    """List components affected by specific CVE."""
    if not cve_id:
        click.echo(ctx.get_help())
        exit(0)
    q = query_service.invoke(core_queries.components_affected_by_specific_cve_query, ctx.params)
    cprint(q, ctx=ctx)


@queries_grp.command(
    name="products-affected-by-flaw",
    help="List Products affected by Flaw.",
)
@click.argument("cve_id", required=False, type=click.STRING, shell_complete=get_cve_ids)
@click.pass_context
@progress_bar
def product_versions_affected_by_cve_query(ctx, cve_id):
    """List Products affected by a CVE."""
    if not cve_id:
        click.echo(ctx.get_help())
        exit(0)
    q = query_service.invoke(
        core_queries.products_versions_affected_by_specific_cve_query, ctx.params
    )
    cprint(q, ctx=ctx)


@queries_grp.command(name="component-flaws", help="List Flaws affecting a Component.")
@click.argument("component_name", required=False)
@click.option("--purl")
@click.option(
    "--flaw-state",
    "flaw_state",
    help="Filter by Flaw state.",
    type=click.Choice(OSIDBService.get_flaw_states()),
)
@click.option(
    "--flaw-impact",
    "flaw_impact",
    help="Filter by Flaw impact.",
    type=click.Choice(OSIDBService.get_flaw_impacts()),
)
@click.option(
    "--flaw-resolution",
    "flaw_resolution",
    help="Filter by Flaw resolution.",
    type=click.Choice(OSIDBService.get_flaw_resolutions()),
)
@click.option(
    "--affectedness",
    help="Filter by Affect affectedness.",
    type=click.Choice(OSIDBService.get_affect_affectedness()),
)
@click.option(
    "--affect-resolution",
    "affect_resolution",
    help="Filter by Affect resolution.",
    type=click.Choice(OSIDBService.get_affect_resolution()),
)
@click.option(
    "--affect-impact",
    "affect_impact",
    help="Filter by Affect impact.",
    type=click.Choice(OSIDBService.get_affect_impact()),
)
@click.option(
    "-s",
    "strict_name_search",
    is_flag=True,
    default=False,
    help="Strict search, exact match of component name.",
)
@click.pass_context
@progress_bar
def cves_for_specific_component_query(
    ctx,
    component_name,
    purl,
    flaw_state,
    flaw_impact,
    flaw_resolution,
    affectedness,
    affect_resolution,
    affect_impact,
    strict_name_search,
):
    """List flaws of a specific component."""
    if not purl and not component_name:
        click.echo(ctx.get_help())
        exit(0)

    q = query_service.invoke(core_queries.cves_for_specific_component_query, ctx.params)
    cprint(q, ctx=ctx)


@queries_grp.command(
    name="product-flaws",
    help="List Flaws affecting a Product.",
)
@click.argument(
    "product_version_name",
    required=False,
    type=click.STRING,
    shell_complete=get_product_version_names,
)
@click.option("--ofuri")
@click.option(
    "--flaw-state",
    "flaw_state",
    help="Filter by Flaw state.",
    type=click.Choice(OSIDBService.get_flaw_states()),
)
@click.option(
    "--flaw-impact",
    "flaw_impact",
    help="Filter by Flaw impact.",
    type=click.Choice(OSIDBService.get_flaw_impacts()),
)
@click.option(
    "--flaw-resolution",
    "flaw_resolution",
    help="Filter by Flaw resolution.",
    type=click.Choice(OSIDBService.get_flaw_resolutions()),
)
@click.option(
    "--affectedness",
    help="Filter by Affect affectedness.",
    type=click.Choice(OSIDBService.get_affect_affectedness()),
)
@click.option(
    "--affect-resolution",
    "affect_resolution",
    help="Filter by Affect resolution.",
    type=click.Choice(OSIDBService.get_affect_resolution()),
)
@click.option(
    "--affect-impact",
    "affect_impact",
    help="Filter by Affect impact.",
    type=click.Choice(OSIDBService.get_affect_impact()),
)
@click.option(
    "-s",
    "strict_name_search",
    is_flag=True,
    default=False,
    help="Strict search, exact match of component name.",
)
@click.pass_context
@progress_bar
def cves_for_specific_product_query(
    ctx,
    product_version_name,
    ofuri,
    flaw_state,
    flaw_impact,
    flaw_resolution,
    affectedness,
    affect_impact,
    affect_resolution,
    strict_name_search,
):
    """List flaws of a specific product."""
    if not product_version_name and not ofuri:
        click.echo(ctx.get_help())
        exit(0)
    q = query_service.invoke(core_queries.cves_for_specific_product_query, ctx.params)
    cprint(q, ctx=ctx)



import logging

import requests

from griffon import CORGI_API_URL, OSIDB_API_URL

logger = logging.getLogger("griffon")


def get_product_version_ofuris(ctx, param, incomplete):
    payload = {"limit": 100, "include_fields": "ofuri", "re_ofuri": incomplete}
    response = requests.get(
        f"{CORGI_API_URL}/api/v1/product_versions",
        params=payload,
        headers={"Accept-Encoding": "gzip;q=1.0, identity; q=0.5, *;q=0"},
    )
    ofuris = response.json()["results"]
    return [k["ofuri"] for k in ofuris if k["ofuri"].startswith(incomplete)]


def get_product_version_names(ctx, param, incomplete):
    payload = {"limit": 100, "include_fields": "name", "re_name": incomplete}
    response = requests.get(
        f"{CORGI_API_URL}/api/v1/product_versions",
        params=payload,
        headers={"Accept-Encoding": "gzip;q=1.0, identity; q=0.5, *;q=0"},
    )
    names = response.json()["results"]
    return [k["name"] for k in names if k["name"].startswith(incomplete)]


def get_product_stream_ofuris(ctx, param, incomplete):
    payload = {"limit": 100, "include_fields": "ofuri", "re_ofuri": incomplete}
    response = requests.get(
        f"{CORGI_API_URL}/api/v1/product_streams",
        params=payload,
        headers={"Accept-Encoding": "gzip;q=1.0, identity; q=0.5, *;q=0"},
    )
    ofuris = response.json()["results"]
    return [k["ofuri"] for k in ofuris if k["ofuri"].startswith(incomplete)]


def get_product_stream_names(ctx, param, incomplete):
    payload = {"limit": 100, "include_fields": "name", "re_name": incomplete}
    response = requests.get(
        f"{CORGI_API_URL}/api/v1/product_streams",
        params=payload,
        headers={"Accept-Encoding": "gzip;q=1.0, identity; q=0.5, *;q=0"},
    )
    names = response.json()["results"]
    return [k["name"] for k in names if k["name"].startswith(incomplete)]


def get_component_names(ctx, param, incomplete):
    payload = {"limit": 100, "include_fields": "name", "re_name": incomplete}
    response = requests.get(
        f"{CORGI_API_URL}/api/v1/components",
        params=payload,
        headers={"Accept-Encoding": "gzip;q=1.0, identity; q=0.5, *;q=0"},
    )
    names = response.json()["results"]
    return list(set([k["name"] for k in names if k["name"].startswith(incomplete)]))


def get_component_purls(ctx, param, incomplete):
    payload = {"limit": 100, "include_fields": "purl", "re_purl": incomplete}
    response = requests.get(
        f"{CORGI_API_URL}/api/v1/components",
        params=payload,
        headers={"Accept-Encoding": "gzip;q=1.0, identity; q=0.5, *;q=0"},
    )
    names = response.json()["results"]
    return list(set([k["purl"] for k in names if k["purl"].startswith(incomplete)]))


def get_cve_ids(ctx, param, incomplete):
    """TODO - the following is not ideal for autocomplete lookup - need to investigate"""
    response = requests.get(
        f"{OSIDB_API_URL}/osidb/api/v1/flaws?limit=10&re_cve_id={incomplete}&include_fields=cve_id"  # noqa
    )
    return [k["cve_id"] for k in response.json()["results"] if k["cve_id"].startswith(incomplete)]



"""
FCC plugin:


"""
import click

api_url = "https://cve.mitre.org"


@click.group()
@click.pass_context
def plugins(ctx):
    """mitre cve plugin"""
    pass


@plugins.command()
@click.argument(
    "cve_id",
    required=True,
    type=click.STRING,
)
@click.pass_context
def retrieve(ctx, cve_id):
    """Retrieve cve mitre by CVE ID"""
    if not cve_id:
        click.echo(ctx.get_help())
        exit(0)
    click.launch(f"{api_url}/cgi-bin/cvename.cgi?name={cve_id}")


@plugins.command()
@click.argument(
    "query_string",
    required=True,
    type=click.STRING,
)
@click.pass_context
def search(ctx, query_string):
    """Search cve mitre by query term"""
    if not query_string:
        click.echo(ctx.get_help())
        exit(0)
    click.launch(f"{api_url}/cgi-bin/cvekey.cgi?keyword={query_string}")
